from slapos.tests.test_slapproxy import BasicMixin
from slapos.util import dumps
import unittest
import json
import mock


class JsonRpcTestCase(BasicMixin, unittest.TestCase):
  #######################################################
  # Get hateoas url
  #######################################################
  def test_post_v0_hateoas_url(self):
    response = self.app.post(
      '/slapos.get.v0.hateoas_url',
      json={}
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        "hateoas_url": "http://localhost/hateoas/"
    }
    assert json.loads(response.data) == expect_result_dict, response.data

  def test_post_v0_hateoas_url_with_https(self):
    response = self.app.post(
      '/slapos.get.v0.hateoas_url',
      json={},
      headers={'X-Forwarded-Proto': 'https'}
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        "hateoas_url": "https://localhost/hateoas/"
    }
    assert json.loads(response.data) == expect_result_dict, response.data

  #######################################################
  # remove compute node certificate
  #######################################################
  def test_remove_v0_compute_node_certificate(self):
    response = self.app.post(
      '/slapos.remove.v0.compute_node_certificate',
      json={
        'computer_guid': 'foo'
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        "type": "success",
        "title": "Ignored"
    }
    assert json.loads(response.data) == expect_result_dict, response.data

  #######################################################
  # post computer certificate
  #######################################################
  def test_post_v0_compute_node_certificate(self):
    response = self.app.post(
      '/slapos.post.v0.compute_node_certificate',
      json={
        'computer_guid': 'foo',
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        "key": "",
        "certificate": ""
    }
    assert json.loads(response.data) == expect_result_dict, response.data

  #######################################################
  # post compute node usage
  #######################################################
  def test_post_v0_compute_node_usage(self):
    response = self.app.post(
      '/slapos.post.v0.compute_node_usage',
      json={
        'computer_guid': 'foo',
        'tioxml': ''
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        "type": "success",
        "title": "Ignored"
    }
    assert json.loads(response.data) == expect_result_dict, response.data

  #######################################################
  # get compute node status
  #######################################################
  def test_get_v0_compute_node_status(self):
    response = self.app.post(
      '/slapos.get.v0.compute_node_status',
      json={
        'computer_guid': self.computer_id,
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        "text": "Unknown (not implemented)"
    }
    assert json.loads(response.data) == expect_result_dict, response.data

  #######################################################
  # get software instance certificate
  #######################################################
  def test_get_v0_software_instance_certificate(self):
    response = self.app.post(
      '/slapos.get.v0.software_instance_certificate',
      json={
        'instance_guid': 'foo',
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        "key": "",
        "certificate": ""
    }
    assert json.loads(response.data) == expect_result_dict, response.data

  #######################################################
  # post software installation
  #######################################################
  def test_post_v0_software_installation(self):
    self.format_for_number_of_partitions(0)
    software_release_url = 'https://mysoft'

    response = self.app.post(
      '/slapos.post.v0.software_installation',
      json={
        'software_release_uri': software_release_url,
        'computer_guid': self.computer_id
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        "type": "success",
        "title": "Supplied https://mysoft to be available"
    }
    assert json.loads(response.data) == expect_result_dict, response.data

  #######################################################
  # put software installation reported state
  #######################################################
  def test_put_v0_software_installation_reported_state_available(self):
    response = self.app.post(
      '/slapos.put.v0.software_installation_reported_state',
      json={
        'software_release_uri': 'foo',
        'computer_guid': 'bar',
        'reported_state': 'available'
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        "type": "success",
        "title": "Ignored"
    }
    assert json.loads(response.data) == expect_result_dict, response.data

  def test_put_v0_software_installation_reported_state_destroyed(self):
    self.format_for_number_of_partitions(0)
    software_release_url = 'https://mysoft'

    self.app.post(
      '/slapos.post.v0.software_installation',
      json={
        'software_release_uri': software_release_url,
        'computer_guid': self.computer_id
      }
    )

    response = self.app.post(
      '/slapos.put.v0.software_installation_reported_state',
      json={
        'software_release_uri': software_release_url,
        'computer_guid': self.computer_id,
        'reported_state': 'destroyed'
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        "type": "success",
        "title": "Destroyed"
    }
    assert json.loads(response.data) == expect_result_dict, response.data

    response = self.app.post(
      '/slapos.allDocs.v0.compute_node_software_installation_list',
      json={
        'computer_guid': self.computer_id
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        "result_list": []
    }
    assert json.loads(response.data) == expect_result_dict, response.data

  #######################################################
  # compute_node_software_installation_list
  #######################################################
  def test_allDocs_v0_compute_node_software_installation_list(self):
    self.format_for_number_of_partitions(0)
    software_release_url = 'https://mysoft'
    response = self.app.post(
      '/slapos.post.v0.software_installation',
      json={
        'software_release_uri': software_release_url,
        'computer_guid': self.computer_id
      }
    )

    response = self.app.post(
      '/slapos.allDocs.v0.compute_node_software_installation_list',
      json={
        'computer_guid': self.computer_id
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        "result_list": [{
          "software_release_uri": software_release_url,
          "state": "available"
        }]
    }
    assert json.loads(response.data) == expect_result_dict, response.data

  #######################################################
  # Software Instance
  #######################################################
  def test_post_v0_software_instance__no_partition(self):
    response = self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'foo',
        'software_release_uri': 'bar',
        'software_type': 'foobar'
      }
    )

    assert response.status_code == 403, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        "status": 403,
        "type": "Forbidden",
        "title": "No free computer partition found for: foo"
    }
    assert json.loads(response.data) == expect_result_dict, response.data

  def test_post_v0_software_instance__first_allocation(self):
    self.format_for_number_of_partitions(1)
    response = self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MyFirstInstance',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar',
        'parameters': {'bar': 'foo'}
      }
    )

    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        'title': 'MyFirstInstance',
        'instance_guid': 'MyFirstInstance______0',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar',
        'state': 'started',
        'connection_parameters': {},
        'parameters': {'bar': 'foo'},
        'shared': False,
        'root_instance_title': 'MyFirstInstance',
        'ip_list': [["tap0", "1.2.3.4"], ["tap0", "4.3.2.1"]],
        'full_ip_list': [],
        'sla_parameters': {},
        'computer_guid': 'computer',
        'compute_partition_id': 'slappart0',
        'processing_timestamp': None,
        'access_status_message': ""
    }
    data_result = json.loads(response.data)
    expect_result_dict['processing_timestamp'] = data_result.get('processing_timestamp', 'unknown')
    assert data_result == expect_result_dict, response.data

  def test_post_v0_software_instance__with_connection_parameters(self):
    self.format_for_number_of_partitions(1)
    response_dict = json.loads(self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MyFirstInstance',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar',
        'parameters': {'bar': 'foo'}
      }
    ).data)

    self.app.post(
      '/slapos.put.v0.software_instance_connection_parameter',
      json={
        'instance_guid': 'MyFirstInstance______0',
        'connection_parameter_dict': {
          'foo': 'bar'
        }
      }
    )

    # Get updated information for the partition
    response = self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MyFirstInstance',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar',
        'parameters': {'bar': 'foo'}
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        'title': 'MyFirstInstance',
        'instance_guid': 'MyFirstInstance______0',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar',
        'state': 'started',
        'connection_parameters': {'foo': 'bar'},
        'parameters': {'bar': 'foo'},
        'shared': False,
        'root_instance_title': 'MyFirstInstance',
        'ip_list': [["tap0", "1.2.3.4"], ["tap0", "4.3.2.1"]],
        'full_ip_list': [],
        'sla_parameters': {},
        'computer_guid': 'computer',
        'compute_partition_id': 'slappart0',
        'processing_timestamp': None,
        'access_status_message': ""
    }
    data_result = json.loads(response.data)
    expect_result_dict['processing_timestamp'] = data_result.get('processing_timestamp', 'unknown')
    assert data_result == expect_result_dict, response.data

  #######################################################
  # Shared Instance
  #######################################################
  def test_post_v0_shared_instance__with_connection_parameters(self):
    # First, create a Software Instance
    self.format_for_number_of_partitions(1)
    self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MyFirstInstance',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar',
        'parameters': {'bar': 'foo'}
      }
    )

    # Second, create a Shared Instance
    response = self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MySharedInstance',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar',
        'parameters': {'bar2': 'foo2'},
        'shared': True
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        'title': 'MySharedInstance',
        'instance_guid': 'MySharedInstance______1',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar',
        'state': 'started',
        'connection_parameters': {},
        'parameters': {'bar2': 'foo2'},
        'shared': True,
        'root_instance_title': 'MySharedInstance',
        'ip_list': [["tap0", "1.2.3.4"], ["tap0", "4.3.2.1"]],
        'full_ip_list': [],
        'sla_parameters': {},
        'computer_guid': 'computer',
        'compute_partition_id': 'slappart0',
        'processing_timestamp': 0,
        'access_status_message': ""
    }
    data_result = json.loads(response.data)
    assert data_result == expect_result_dict, response.data


    self.app.post(
      '/slapos.put.v0.software_instance_connection_parameter',
      json={
        'instance_guid': 'MySharedInstance______1',
        'connection_parameter_dict': {
          'foo': 'bar'
        }
      }
    )

    # Get updated information for the partition
    response = self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MySharedInstance',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar',
        'parameters': {'bar2': 'foo2'},
        'shared': True
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        'title': 'MySharedInstance',
        'instance_guid': 'MySharedInstance______1',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar',
        'state': 'started',
        'connection_parameters': {'foo': 'bar'},
        'parameters': {'bar2': 'foo2'},
        'shared': True,
        'root_instance_title': 'MySharedInstance',
        'ip_list': [["tap0", "1.2.3.4"], ["tap0", "4.3.2.1"]],
        'full_ip_list': [],
        'sla_parameters': {},
        'computer_guid': 'computer',
        'compute_partition_id': 'slappart0',
        'processing_timestamp': 0,
        'access_status_message': ""
    }
    data_result = json.loads(response.data)
    expect_result_dict['processing_timestamp'] = data_result.get('processing_timestamp', 'unknown')
    assert data_result == expect_result_dict, response.data

  #######################################################
  # CDN Shared Instance
  #######################################################
  def test_post_v0_shared_instance__with_cdn(self):
    response = self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MyCDNInstance',
        'software_release_uri': 'http://git.erp5.org/gitweb/slapos.git/blob_plain/HEAD:/software/apache-frontend/software.cfg',
        'software_type': 'default',
        'parameters': {'url': 'https://[::1]:123/my/path?my=query&string=value#myanchor'},
        'shared': True
      }
    )

    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        'title': 'MyCDNInstance',
        'instance_guid': 'MyCDNInstance______1',
        'software_release_uri': 'http://git.erp5.org/gitweb/slapos.git/blob_plain/HEAD:/software/apache-frontend/software.cfg',
        'software_type': 'default',
        'state': 'started',
        'connection_parameters': {
          'secure_access': 'https://[::1]:123/my/path?my=query&string=value#myanchor',
          'domain': '[::1]:123'
        },
        'parameters': {'url': 'https://[::1]:123/my/path?my=query&string=value#myanchor'},
        'shared': True,
        'root_instance_title': 'MyCDNInstance',
        'ip_list': [],
        'full_ip_list': [],
        'sla_parameters': {},
        'computer_guid': '',
        'compute_partition_id': 'Fake frontend for https://[::1]:123/my/path?my=query&string=value#myanchor',
        'processing_timestamp': 0,
        'access_status_message': ""
    }
    data_result = json.loads(response.data)
    assert data_result == expect_result_dict, response.data

  def test_post_v0_shared_instance__with_cdn_and_https_proxy(self):
    response = self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MyCDNInstance',
        'software_release_uri': 'http://git.erp5.org/gitweb/slapos.git/blob_plain/HEAD:/software/apache-frontend/software.cfg',
        'software_type': 'default',
        'parameters': {'url': 'https://[::1]:123/my/path?my=query&string=value#myanchor'},
        'shared': True
      },
      headers={'X-Forwarded-Proto': 'https'}
    )

    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        'title': 'MyCDNInstance',
        'instance_guid': 'MyCDNInstance______1',
        'software_release_uri': 'http://git.erp5.org/gitweb/slapos.git/blob_plain/HEAD:/software/apache-frontend/software.cfg',
        'software_type': 'default',
        'state': 'started',
        'connection_parameters': {
          'secure_access': 'https://localhost/http_proxy/https/%5B::1%5D:123/my/path?my=query&string=value#myanchor',
          'domain': 'localhost'
        },
        'parameters': {'url': 'https://[::1]:123/my/path?my=query&string=value#myanchor'},
        'shared': True,
        'root_instance_title': 'MyCDNInstance',
        'ip_list': [],
        'full_ip_list': [],
        'sla_parameters': {},
        'computer_guid': '',
        'compute_partition_id': 'Fake frontend for https://[::1]:123/my/path?my=query&string=value#myanchor',
        'processing_timestamp': 0,
        'access_status_message': ""
    }
    data_result = json.loads(response.data)
    assert data_result == expect_result_dict, response.data

  #######################################################
  # slapos.allDocs.v0.compute_node_instance_list
  #######################################################
  def test_allDocs_v0_compute_node_instance_list(self):
    self.format_for_number_of_partitions(1)
    self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MyFirstInstance',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar',
        'parameters': {'bar': 'foo'}
      }
    )

    response = self.app.post(
      '/slapos.allDocs.v0.compute_node_instance_list',
      json={
        'computer_guid': self.computer_id
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        'result_list': [{
          'title': 'MyFirstInstance',
          'instance_guid': 'MyFirstInstance______0',
          'state': 'started',
          'compute_partition_id': 'slappart0',
          'software_release_uri': 'http://sr//'
        }]
    }
    data_result = json.loads(response.data)
    assert data_result == expect_result_dict, response.data

  #######################################################
  # slapos.allDocs.v0.instance_node_instance_list
  #######################################################
  def test_allDocs_v0_instance_node_instance_list_no_instance(self):
    response = self.app.post(
      '/slapos.allDocs.v0.instance_node_instance_list',
      json={
        'instance_guid': 'MyFirstInstance______0'
      }
    )
    assert response.status_code == 403, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        'status': 403,
        'type': 'Forbidden',
        'title': "No software instance MyFirstInstance______0 found."
    }
    data_result = json.loads(response.data)
    assert data_result == expect_result_dict, response.data

  def test_allDocs_v0_instance_node_instance_list_empty(self):
    self.format_for_number_of_partitions(1)
    self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MyFirstInstance',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar',
        'parameters': {'bar': 'foo'}
      }
    )

    response = self.app.post(
      '/slapos.allDocs.v0.instance_node_instance_list',
      json={
        'instance_guid': 'MyFirstInstance______0'
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        'result_list': []
    }
    data_result = json.loads(response.data)
    assert data_result == expect_result_dict, response.data

  def test_allDocs_v0_instance_node_instance_list_not_empty(self):
    self.format_for_number_of_partitions(1)
    self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MyFirstInstance',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar',
        'parameters': {'bar': 'foo'}
      }
    )
    self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MySharedInstance',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar',
        'parameters': {'bar2': 'foo2'},
        'shared': True
      }
    )

    response = self.app.post(
      '/slapos.allDocs.v0.instance_node_instance_list',
      json={
        'instance_guid': 'MyFirstInstance______0'
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        'result_list': [{
          "title": "MySharedInstance",
          "instance_guid": "MySharedInstance______1",
          "software_type": "foobar",
          "state": "started",
          "parameters": {'bar2': 'foo2'},
          "compute_partition_id": "slappart0"
        }]
    }
    data_result = json.loads(response.data)
    assert data_result == expect_result_dict, response.data

  #######################################################
  # slapos.get.v0.software_instance
  #######################################################
  def test_get_v0_computer_partition__not_instance(self):
    response = self.app.post(
      '/slapos.get.v0.compute_partition',
      json={
        'computer_guid': 'foo',
        'compute_partition_id': 'bar'
      }
    )
    assert response.status_code == 403, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        'status': 403,
        'type': 'Forbidden',
        'title': 'No instance on partition bar found.'
    }
    data_result = json.loads(response.data)
    assert data_result == expect_result_dict, response.data

  def test_get_v0_computer_partition__matching_instance(self):
    self.format_for_number_of_partitions(1)
    response_dict = json.loads(self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MyFirstInstance',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar',
        'parameters': {'bar': 'foo'}
      }
    ).data)

    self.app.post(
      '/slapos.put.v0.software_instance_connection_parameter',
      json={
        'instance_guid': 'MyFirstInstance______0',
        'connection_parameter_dict': {
          'foo': 'bar'
        }
      }
    )

    # Get updated information for the partition
    response = self.app.post(
      '/slapos.get.v0.compute_partition',
      json={
        'computer_guid': 'computer',
        'compute_partition_id': 'slappart0'
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        'title': 'MyFirstInstance',
        'instance_guid': 'MyFirstInstance______0',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar',
        'state': 'started',
        'connection_parameters': {'foo': 'bar'},
        'parameters': {'bar': 'foo'},
        'shared': False,
        'root_instance_title': 'MyFirstInstance',
        'ip_list': [["tap0", "1.2.3.4"], ["tap0", "4.3.2.1"]],
        'full_ip_list': [],
        'sla_parameters': {},
        'computer_guid': 'computer',
        'compute_partition_id': 'slappart0',
        'processing_timestamp': None,
        'access_status_message': ""
    }
    data_result = json.loads(response.data)
    expect_result_dict['processing_timestamp'] = data_result.get('processing_timestamp', 'unknown')
    assert data_result == expect_result_dict, response.data

  def test_get_v0_computer_partition__empty_partition(self):
    self.format_for_number_of_partitions(1)

    # Get information for the partition
    response = self.app.post(
      '/slapos.get.v0.compute_partition',
      json={
        'computer_guid': 'computer',
        'compute_partition_id': 'slappart0'
      }
    )
    assert response.status_code == 403, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        'status': 403,
        'type': 'Forbidden',
        'title': 'No instance on partition slappart0 found.'
    }
    data_result = json.loads(response.data)
    assert data_result == expect_result_dict, response.data

  #######################################################
  # slapos.get.v0.software_instance
  #######################################################
  def test_get_v0_software_instance__not_instance(self):
    response = self.app.post(
      '/slapos.get.v0.software_instance',
      json={
        'instance_guid': 'foo'
      }
    )
    assert response.status_code == 403, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        'status': 403,
        'type': 'Forbidden',
        'title': 'instance_guid foo not handled.'
    }
    data_result = json.loads(response.data)
    assert data_result == expect_result_dict, response.data

  def test_get_v0_software_instance__matching_instance(self):
    self.format_for_number_of_partitions(1)
    response_dict = json.loads(self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MyFirstInstance',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar',
        'parameters': {'bar': 'foo'}
      }
    ).data)

    self.app.post(
      '/slapos.put.v0.software_instance_connection_parameter',
      json={
        'instance_guid': 'MyFirstInstance______0',
        'connection_parameter_dict': {
          'foo': 'bar'
        }
      }
    )

    # Get updated information for the partition
    response = self.app.post(
      '/slapos.get.v0.software_instance',
      json={
        'instance_guid': 'MyFirstInstance______0'
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        'title': 'MyFirstInstance',
        'instance_guid': 'MyFirstInstance______0',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar',
        'state': 'started',
        'connection_parameters': {'foo': 'bar'},
        'parameters': {'bar': 'foo'},
        'shared': False,
        'root_instance_title': 'MyFirstInstance',
        'ip_list': [["tap0", "1.2.3.4"], ["tap0", "4.3.2.1"]],
        'full_ip_list': [],
        'sla_parameters': {},
        'computer_guid': 'computer',
        'compute_partition_id': 'slappart0',
        'processing_timestamp': None,
        'access_status_message': ""
    }
    data_result = json.loads(response.data)
    expect_result_dict['processing_timestamp'] = data_result.get('processing_timestamp', 'unknown')
    assert data_result == expect_result_dict, response.data

  def test_get_v0_software_instance__matching_shared_instance(self):
    self.format_for_number_of_partitions(1)
    self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MyFirstInstance',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar'
      }
    )
    response_dict = json.loads(self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MyFirstShared',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar',
        'shared': True,
        'parameters': {'bar': 'foo'}
      }
    ).data)

    self.app.post(
      '/slapos.put.v0.software_instance_connection_parameter',
      json={
        'instance_guid': 'MyFirstShared______1',
        'connection_parameter_dict': {
          'foo': 'bar'
        }
      }
    )

    # Get updated information for the partition
    response = self.app.post(
      '/slapos.get.v0.software_instance',
      json={
        'instance_guid': 'MyFirstShared______1'
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        'title': 'MyFirstShared',
        'instance_guid': 'MyFirstShared______1',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar',
        'state': 'started',
        'connection_parameters': {'foo': 'bar'},
        'parameters': {'bar': 'foo'},
        'shared': True,
        'root_instance_title': 'MyFirstShared',
        'ip_list': [],
        'full_ip_list': [],
        'sla_parameters': {},
        'computer_guid': 'computer',
        'compute_partition_id': 'slappart0',
        'processing_timestamp': None,
        'access_status_message': ""
    }
    data_result = json.loads(response.data)
    expect_result_dict['processing_timestamp'] = data_result.get('processing_timestamp', 'unknown')
    assert data_result == expect_result_dict, response.data

  #######################################################
  # put compute node format
  #######################################################
  def test_put_v0_compute_node_format_empty_node(self):
    response = self.app.post(
      '/slapos.put.v0.compute_node_format',
      json={
        'computer_guid': self.computer_id,
        'compute_partition_list': []
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        "type": "success",
        "title": "Formatted"
    }
    assert json.loads(response.data) == expect_result_dict, response.data

    # No partition should be available for an instance request
    software_release_url = 'https://mysoft'
    response = self.app.post(
      '/slapos.post.v0.software_installation',
      json={
        'software_release_uri': software_release_url,
        'computer_guid': self.computer_id
      }
    )
    assert response.status_code == 200, response.status_code

    response = self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MyFirstInstance',
        'software_release_uri': software_release_url,
        'software_type': 'foobar'
      }
    )
    assert response.status_code == 403, response.status_code
    expect_result_dict = {
        "status": 403,
        "type": "Forbidden",
        "title": "No free computer partition found for: MyFirstInstance"
    }
    assert json.loads(response.data) == expect_result_dict, response.data

  def test_put_v0_compute_node_format_new_partition(self):
    response = self.app.post(
      '/slapos.put.v0.compute_node_format',
      json={
        'computer_guid': self.computer_id,
        'compute_partition_list': [{
          'partition_id': 'MyFirstPartition',
          'ip_list': [{
            'ip-address': 'MyFirstIpAddress',
            'network-interface': 'MyFirstNetworkInterface',
            'network-address': 'MyFirstNetworkAddress',
            'gateway-ip-address': 'MyFirstGatewayIpAddress',
            'netmask': 'MyFirstNetmask',
          }, {
            'ip-address': 'MySecondIpAddress',
            'network-interface': 'MySecondNetworkInterface',
            'network-address': 'MySecondNetworkAddress',
            'gateway-ip-address': 'MySecondGatewayIpAddress',
            'netmask': 'MySecondNetmask',
          }]
        }]
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        "type": "success",
        "title": "Formatted"
    }
    assert json.loads(response.data) == expect_result_dict, response.data

    # A partition should be available for an instance request
    software_release_url = 'https://mysoft'
    response = self.app.post(
      '/slapos.post.v0.software_installation',
      json={
        'software_release_uri': software_release_url,
        'computer_guid': self.computer_id
      }
    )
    assert response.status_code == 200, response.status_code

    response = self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MyFirstInstance',
        'software_release_uri': software_release_url,
        'software_type': 'foobar'
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        'title': 'MyFirstInstance',
        'instance_guid': 'MyFirstInstance______0',
        'software_release_uri': software_release_url,
        'software_type': 'foobar',
        'state': 'started',
        'connection_parameters': {},
        'parameters': {},
        'shared': False,
        'root_instance_title': 'MyFirstInstance',
        'ip_list': [
          ["MyFirstNetworkInterface", "MyFirstIpAddress"],
          ["MySecondNetworkInterface", "MySecondIpAddress"]
        ],
        'full_ip_list': [],
        'sla_parameters': {},
        'computer_guid': 'computer',
        'compute_partition_id': 'MyFirstPartition',
        'processing_timestamp': None,
        'access_status_message': ""
    }
    data_result = json.loads(response.data)
    expect_result_dict['processing_timestamp'] = data_result.get('processing_timestamp', 'unknown')
    assert data_result == expect_result_dict, response.data

  def test_put_v0_compute_node_format_update_existing(self):
    # First, create 2 partitions
    response = self.app.post(
      '/slapos.put.v0.compute_node_format',
      json={
        'computer_guid': self.computer_id,
        'compute_partition_list': [{
          'partition_id': 'MyFirstPartition',
          'ip_list': [{
            'ip-address': 'MyFirstIpAddress',
            'network-interface': 'MyFirstNetworkInterface',
            'network-address': 'MyFirstNetworkAddress',
            'gateway-ip-address': 'MyFirstGatewayIpAddress',
            'netmask': 'MyFirstNetmask',
          }]
        }, {
          'partition_id': 'MySecondPartition',
          'ip_list': [{
            'ip-address': 'MySecondIpAddress',
            'network-interface': 'MySecondNetworkInterface',
            'network-address': 'MySecondNetworkAddress',
            'gateway-ip-address': 'MySecondGatewayIpAddress',
            'netmask': 'MySecondNetmask',
          }]
        }]
      }
    )
    assert response.status_code == 200, response.status_code

    # Second, install 1 SR
    software_release_url = 'https://mysoft1'
    response = self.app.post(
      '/slapos.post.v0.software_installation',
      json={
        'software_release_uri': software_release_url,
        'computer_guid': self.computer_id
      }
    )
    assert response.status_code == 200, response.status_code

    # Third, allocate 2 software instances and 2 slave instances
    software_type1 = 'foobar1'
    software_type2 = 'foobar2'
    response = self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MyFirstInstance',
        'software_release_uri': software_release_url,
        'software_type': software_type1
      }
    )
    assert response.status_code == 200, response.status_code
    response = self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MyFirstShared',
        'software_release_uri': software_release_url,
        'software_type': software_type1,
        'shared': True
      }
    )
    assert response.status_code == 200, response.status_code
    response = self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MySecondInstance',
        'software_release_uri': software_release_url,
        'software_type': software_type2
      }
    )
    assert response.status_code == 200, response.status_code
    response = self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MySecondShared',
        'software_release_uri': software_release_url,
        'software_type': software_type2,
        'shared': True
      }
    )
    assert response.status_code == 200, response.status_code

    # And now, time to update the format
    # keep 1 partition with updated ip, drop 1 partition, and add another one
    response = self.app.post(
      '/slapos.put.v0.compute_node_format',
      json={
        'computer_guid': self.computer_id,
        'compute_partition_list': [{
          'partition_id': 'MyThirdPartition',
          'ip_list': [{
            'ip-address': 'MyThirdIpAddress',
            'network-interface': 'MyThirdNetworkInterface',
            'network-address': 'MyThirdNetworkAddress',
            'gateway-ip-address': 'MyThirdGatewayIpAddress',
            'netmask': 'MyThirdNetmask',
          }]
        }, {
          'partition_id': 'MySecondPartition',
          'ip_list': [{
            'ip-address': 'MyNewSecondIpAddress',
            'network-interface': 'MyNewSecondNetworkInterface',
            'network-address': 'MyNewSecondNetworkAddress',
            'gateway-ip-address': 'MyNewSecondGatewayIpAddress',
            'netmask': 'MyNewSecondNetmask',
          }]
        }]
      }
    )
    assert response.status_code == 200, response.status_code

    # Ensure the 2 instances on MySecondPartition were kept
    response = self.app.post(
      '/slapos.get.v0.software_instance',
      json={
        'instance_guid': 'MySecondInstance______0'
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        'title': 'MySecondInstance',
        'instance_guid': 'MySecondInstance______0',
        'software_release_uri': software_release_url,
        'software_type': software_type2,
        'state': 'started',
        'connection_parameters': {},
        'parameters': {},
        'shared': False,
        'root_instance_title': 'MySecondInstance',
        'ip_list': [["MyNewSecondNetworkInterface", "MyNewSecondIpAddress"]],
        'full_ip_list': [],
        'sla_parameters': {},
        'computer_guid': 'computer',
        'compute_partition_id': 'MySecondPartition',
        'processing_timestamp': None,
        'access_status_message': ""
    }
    data_result = json.loads(response.data)
    expect_result_dict['processing_timestamp'] = data_result.get('processing_timestamp', 'unknown')
    assert data_result == expect_result_dict, response.data

    response = self.app.post(
      '/slapos.get.v0.software_instance',
      json={
        'instance_guid': 'MySecondShared______1'
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        'title': 'MySecondShared',
        'instance_guid': 'MySecondShared______1',
        'software_release_uri': software_release_url,
        'software_type': software_type2,
        'state': 'started',
        'connection_parameters': {},
        'parameters': {},
        'shared': True,
        'root_instance_title': 'MySecondShared',
        'ip_list': [],
        'full_ip_list': [],
        'sla_parameters': {},
        'computer_guid': 'computer',
        'compute_partition_id': 'MySecondPartition',
        'processing_timestamp': None,
        'access_status_message': ""
    }
    data_result = json.loads(response.data)
    expect_result_dict['processing_timestamp'] = data_result.get('processing_timestamp', 'unknown')
    assert data_result == expect_result_dict, response.data

    # Check the instances on MyFirstPartition were dropped
    # (this is a different behaviour then erp5 currently)
    response = self.app.post(
      '/slapos.get.v0.software_instance',
      json={
        'instance_guid': 'MyFirstInstance______0'
      }
    )
    assert response.status_code == 403, response.status_code
    response = self.app.post(
      '/slapos.get.v0.software_instance',
      json={
        'instance_guid': 'MyFirstShared______1'
      }
    )
    assert response.status_code == 403, response.status_code

    # And finally, we should have 1 last free instance
    response = self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MyThirdInstance',
        'software_release_uri': software_release_url,
        'software_type': 'foobar'
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        'title': 'MyThirdInstance',
        'instance_guid': 'MyThirdInstance______0',
        'software_release_uri': software_release_url,
        'software_type': 'foobar',
        'state': 'started',
        'connection_parameters': {},
        'parameters': {},
        'shared': False,
        'root_instance_title': 'MyThirdInstance',
        'ip_list': [
          ["MyThirdNetworkInterface", "MyThirdIpAddress"]
        ],
        'full_ip_list': [],
        'sla_parameters': {},
        'computer_guid': 'computer',
        'compute_partition_id': 'MyThirdPartition',
        'processing_timestamp': None,
        'access_status_message': ""
    }
    data_result = json.loads(response.data)
    expect_result_dict['processing_timestamp'] = data_result.get('processing_timestamp', 'unknown')
    assert data_result == expect_result_dict, response.data


  #######################################################
  # put compute node bang
  #######################################################
  def test_put_v0_compute_node_bang(self):
    response = self.app.post(
      '/slapos.put.v0.compute_node_bang',
      json={
        'computer_guid': self.computer_id,
        'message': 'it does not work'
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        "type": "success",
        "title": "Ignored"
    }
    assert json.loads(response.data) == expect_result_dict, response.data

  #######################################################
  # put software connection reported state
  #######################################################
  def test_put_v0_software_instance_reported_state_started(self):
    response = self.app.post(
      '/slapos.put.v0.software_instance_reported_state',
      json={
        'instance_guid': 'foo',
        'reported_state': 'started'
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        "type": "success",
        "title": "Ignored"
    }
    assert json.loads(response.data) == expect_result_dict, response.data

  def test_put_v0_software_instance_reported_state_stopped(self):
    response = self.app.post(
      '/slapos.put.v0.software_instance_reported_state',
      json={
        'instance_guid': 'foo',
        'reported_state': 'stopped'
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        "type": "success",
        "title": "Ignored"
    }
    assert json.loads(response.data) == expect_result_dict, response.data

  def test_put_v0_software_instance_reported_state_destroyed(self):
    self.format_for_number_of_partitions(1)
    self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MyFirstInstance',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar'
      }
    )

    response = self.app.post(
      '/slapos.put.v0.software_instance_reported_state',
      json={
        'instance_guid': 'MyFirstInstance______0',
        'reported_state': 'destroyed'
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        "type": "success",
        "title": "Destroyed"
    }
    assert json.loads(response.data) == expect_result_dict, response.data

    # Check that instance was dropped
    response = self.app.post(
      '/slapos.get.v0.software_instance',
      json={
        'instance_guid': 'MyFirstInstance______0'
      }
    )
    assert response.status_code == 403, response.status_code
    expect_result_dict = {
        "status": 403,
        "type": "Forbidden",
        "title": "No software instance MyFirstInstance______0 found."
    }
    assert json.loads(response.data) == expect_result_dict, response.data

  #######################################################
  # put software connection bang
  #######################################################
  def test_put_v0_software_instance_bang_instance(self):
    self.format_for_number_of_partitions(1)
    self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MyFirstInstance',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar'
      }
    )

    # Get previous timestamp
    response = self.app.post(
      '/slapos.get.v0.software_instance',
      json={
        'instance_guid': 'MyFirstInstance______0'
      }
    )
    assert response.status_code == 200, response.status_code
    previous_timestamp = json.loads(response.data).get('processing_timestamp', 'unknown')

    # The API timestamp is rounded by second
    with mock.patch('time.time', return_value=previous_timestamp + 1):
      response = self.app.post(
        '/slapos.put.v0.software_instance_bang',
        json={
          'instance_guid': 'MyFirstInstance______0',
          'message': 'Please reprocess'
        }
      )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        "type": "success",
        "title": "Bang handled"
    }
    assert json.loads(response.data) == expect_result_dict, response.data

    # Check that timestamp changed
    response = self.app.post(
      '/slapos.get.v0.software_instance',
      json={
        'instance_guid': 'MyFirstInstance______0'
      }
    )
    assert response.status_code == 200, response.status_code
    assert previous_timestamp != json.loads(response.data).get('processing_timestamp', 'unknown')

  def test_put_v0_software_instance_bang_shared(self):
    self.format_for_number_of_partitions(1)
    self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MyFirstInstance',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar'
      }
    )
    self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MyFirstShared',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar',
        'shared': True
      }
    )

    response = self.app.post(
      '/slapos.put.v0.software_instance_bang',
      json={
        'instance_guid': 'MyFirstShared______1',
        'message': 'Please reprocess'
      }
    )
    assert response.status_code == 403, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        "status": 403,
        "type": "Forbidden",
        "title": "NotImplemented"
    }
    assert json.loads(response.data) == expect_result_dict, response.data

  #######################################################
  # put software connection title
  #######################################################
  def test_put_v0_software_instance_title_instance(self):
    self.format_for_number_of_partitions(1)
    self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MyFirstInstance',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar'
      }
    )

    response = self.app.post(
      '/slapos.put.v0.software_instance_title',
      json={
        'instance_guid': 'MyFirstInstance______0',
        'title': 'MyRenamedFirstInstance'
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        "type": "success",
        "title": "Renamed"
    }
    assert json.loads(response.data) == expect_result_dict, response.data

    # Check that the new uid is usable
    response = self.app.post(
      '/slapos.get.v0.software_instance',
      json={
        'instance_guid': 'MyRenamedFirstInstance______0'
      }
    )
    assert response.status_code == 200, response.status_code

    # Check that the old uid is not usable
    response = self.app.post(
      '/slapos.get.v0.software_instance',
      json={
        'instance_guid': 'MyFirstInstance______0'
      }
    )
    assert response.status_code == 403, response.status_code

  def test_put_v0_software_instance_title_shared(self):
    self.format_for_number_of_partitions(1)
    self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MyFirstInstance',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar'
      }
    )
    self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MyFirstShared',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar',
        'shared': True
      }
    )

    response = self.app.post(
      '/slapos.put.v0.software_instance_title',
      json={
        'instance_guid': 'MyFirstShared______1',
        'title': 'MyRenamedFirstShared'
      }
    )
    assert response.status_code == 403, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
      "status": 403,
      "type": "Forbidden",
      "title": "NotImplemented"
    }
    assert json.loads(response.data) == expect_result_dict, response.data

    """
    # Check that the new uid is usable
    response = self.app.post(
      '/slapos.get.v0.software_instance',
      json={
        'instance_guid': 'MyRenamedFirstShared______1'
      }
    )
    assert response.status_code == 200, response.status_code

    # Check that the old uid is not usable
    response = self.app.post(
      '/slapos.get.v0.software_instance',
      json={
        'instance_guid': 'MyFirstShared______1'
      }
    )
    assert response.status_code == 403, response.status_code
    """

  #######################################################
  # put software connection parameter
  #######################################################
  def test_put_v0_software_instance_connection_parameter_instance(self):
    self.format_for_number_of_partitions(1)
    self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MyFirstInstance',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar'
      }
    )

    response = self.app.post(
      '/slapos.put.v0.software_instance_connection_parameter',
      json={
        'instance_guid': 'MyFirstInstance______0',
        'connection_parameter_dict': {
          'foo': 'bar',
          'bar': 'foo'
        }
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        "type": "success",
        "title": "Updated"
    }
    assert json.loads(response.data) == expect_result_dict, response.data

  def test_put_v0_software_instance_connection_parameter_shared(self):
    self.format_for_number_of_partitions(1)
    self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MyFirstInstance',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar'
      }
    )
    self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MyFirstShared',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar',
        'shared': True
      }
    )

    response = self.app.post(
      '/slapos.put.v0.software_instance_connection_parameter',
      json={
        'instance_guid': 'MyFirstShared______1',
        'connection_parameter_dict': {
          'foo': 'bar',
          'bar': 'foo'
        }
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        "type": "success",
        "title": "Updated"
    }
    assert json.loads(response.data) == expect_result_dict, response.data

  #######################################################
  # put software instance error
  #######################################################
  def test_put_v0_software_instance_error(self):
    response = self.app.post(
      '/slapos.put.v0.software_instance_error',
      json={
        'instance_guid': 'foo',
        'message': 'it does not work'
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        "type": "success",
        "title": "Ignored"
    }
    assert json.loads(response.data) == expect_result_dict, response.data

  #######################################################
  # put software installation error
  #######################################################
  def test_put_v0_software_installation_error(self):
    software_release_url = 'https://mysoft'

    response = self.app.post(
      '/slapos.put.v0.software_installation_error',
      json={
        'software_release_uri': software_release_url,
        'computer_guid': self.computer_id,
        'message': 'it does not work'
      }
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        "type": "success",
        "title": "Ignored"
    }
    assert json.loads(response.data) == expect_result_dict, response.data


class JsonRpcExperimentalTestCase(BasicMixin, unittest.TestCase):
  #######################################################
  # Get instance tree list
  #######################################################
  def test_allDocs_WIP_instance_tree_list(self):
    self.format_for_number_of_partitions(1)
    self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MyFirstInstance',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar'
      }
    )
    self.app.post(
      '/slapos.post.v0.software_instance',
      json={
        'title': 'MyFirstShared',
        'software_release_uri': 'http://sr//',
        'software_type': 'foobar',
        'shared': True
      }
    )

    response = self.app.post(
      '/slapos.allDocs.WIP.instance_tree_list',
      json={}
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        'result_list': [{
          'title': 'MyFirstInstance',
          'instance_guid': 'MyFirstInstance______0'
        }, {
          'title': 'MyFirstShared',
          'instance_guid': 'MyFirstShared______1'
        }]
    }
    assert json.loads(response.data) == expect_result_dict, response.data

  #######################################################
  # Get compute node list
  #######################################################
  def test_allDocs_WIP_compute_node_list(self):
    self.format_for_number_of_partitions(1)

    response = self.app.post(
      '/slapos.allDocs.WIP.compute_node_list',
      json={}
    )
    assert response.status_code == 200, response.status_code
    assert response.content_type == 'application/json', \
        response.content_type
    expect_result_dict = {
        'result_list': [{
          'computer_guid': 'computer'
        }]
    }
    assert json.loads(response.data) == expect_result_dict, response.data



# coding: utf-8
import logging
import os
import os.path
import pwd
import time

from zope.interface import implementer
from slapos.manager import interface

logger = logging.getLogger(__name__)

@implementer(interface.IManager)
class Manager(object):
  """Manage cgroup's cpuset in terms on initializing and runtime operations.

  CPUSET manager moves PIDs between CPU cores using Linux cgroup system.

  In order to use this feature put "cpuset" into "manager_list" into your slapos
  configuration file inside [slapos] section.

  TODO: there is no limit on number of reserved cores per user.
  """
  cpu_exclusive_file = ".slapos-cpu-exclusive"
  cpuset_path = "/sys/fs/cgroup/cpuset/"
  task_write_mode = "wt"
  config_power_user_option = "power_user_list"

  def __init__(self, config):
    """Retain access to dict-like configuration."""
    self.config = config

  def software(self, software):
    """We don't need to mingle with software."""
    pass

  def softwareTearDown(self, software):
    """We don't need to mingle with software."""
    pass

  def format(self, computer):
    """Create cgroup folder per-CPU with exclusive access to the CPU.

    - Those folders are "/sys/fs/cgroup/cpuset/cpu<N>".
    """
    if not os.path.exists(os.path.join(self.cpuset_path, "cpuset.cpus")):
      logger.warning("CPUSet Manager cannot format computer because cgroups do not exist.")
      return

    for cpu in self._cpu_id_list():
      cpu_path = self._prepare_folder(
        os.path.join(self.cpuset_path, "cpu" + str(cpu)))
      with open(cpu_path + "/cpuset.cpus", "wt") as fx:
        fx.write(str(cpu))  # this cgroup manages only this cpu
      with open(cpu_path + "/cpuset.cpu_exclusive", "wt") as fx:
        fx.write("1")  # manages it exclusively
      with open(cpu_path + "/cpuset.mems", "wt") as fx:
        fx.write("0")  # it doesn't work without that

  def formatTearDown(self, computer):
    pass

  def instance(self, partition):
    """Control runtime state of the computer."""

    if not os.path.exists(os.path.join(self.cpuset_path, "cpu0")):
      # check whether the computer was formatted
      logger.warning("CGROUP's CPUSET Manager cannot update computer because it is not cpuset-formatted.")
      return

    request_file = os.path.join(partition.instance_path, self.cpu_exclusive_file)
    if not os.path.exists(request_file) or not read_file(request_file):
      # This instance does not ask for cpu exclusive access
      return

    # Gather list of users allowed to request exlusive cores
    power_user_list = self.config.get(self.config_power_user_option, "").split()
    uid, gid = partition.getUserGroupId()
    uname = pwd.getpwuid(uid).pw_name
    if uname not in power_user_list:
      logger.warning("User {} not allowed to modify cpuset! "
                     "Allowed users are in {} option in config file.".format(
                        uname, self.config_power_user_option))
      return

    # prepare paths to tasks file for all and per-cpu
    tasks_file = os.path.join(self.cpuset_path, "tasks")
    cpu_tasks_file_list = [os.path.join(cpu_folder, "tasks")
                           for cpu_folder in self._cpu_folder_list()]

    # Gather exclusive CPU usage map {username: set[cpu_id]}
    # We do not need to gather that since we have no limits yet
    #cpu_usage = defaultdict(set)
    #for cpu_id in self._cpu_id_list()[1:]:  # skip the first public CPU
    #  pids = [int(pid)
    #          for pid in read_file(cpu_tasks_file_list[cpu_id]).splitlines()]
    #  for pid in pids:
    #    process = psutil.Process(pid)
    #    cpu_usage[process.username()].add(cpu_id)

    # Move all PIDs from the pool of all CPUs onto the first exclusive CPU.
    running_list = sorted(list(map(int, read_file(tasks_file).split())), reverse=True)
    first_cpu = self._cpu_id_list()[0]
    success_set, refused_set = set(), set()
    for pid in running_list:
      try:
        self._move_task(pid, first_cpu)
        success_set.add(pid)
        time.sleep(0.01)
      except IOError as e:
        refused_set.add(pid)
    logger.debug("Refused to move {:d} PIDs: {!s}\n"
                 "Suceeded in moving {:d} PIDs {!s}\n".format(
                     len(refused_set), refused_set, len(success_set), success_set))

    cpu_folder_list = self._cpu_folder_list()
    generic_cpu_path = cpu_folder_list[0]
    exclusive_cpu_path_list = cpu_folder_list[1:]

    # Gather all running PIDs for filtering out stale PIDs
    running_pid_set = set(running_list)
    running_pid_set.update(map(int, read_file(cpu_tasks_file_list[0]).split()))

    # gather already exclusively running PIDs
    exclusive_pid_set = set()
    for cpu_tasks_file in cpu_tasks_file_list[1:]:
      exclusive_pid_set.update(map(int, read_file(cpu_tasks_file).split()))

    # Move processes to their demanded exclusive CPUs
    with open(request_file, "rt") as fi:
      # take such PIDs which are either really running or are not already exclusive
      request_pid_list = [int(pid) for pid in fi.read().split()
                          if int(pid) in running_pid_set or int(pid) not in exclusive_pid_set]
    with open(request_file, "wt") as fo:
      fo.write("")  # empty file (we will write back only PIDs which weren't moved)
    for request_pid in request_pid_list:
      assigned_cpu = self._move_to_exclusive_cpu(request_pid)
      if assigned_cpu < 0:
        # if no exclusive CPU was assigned - write the PID back and try other time
        with open(request_file, "at") as fo:
          fo.write(str(request_pid) + "\n")

  def instanceTearDown(self, instance):
    pass

  def report(self, partition):
    pass

  def _cpu_folder_list(self):
    """Return list of folders for exclusive cpu cores."""
    return [os.path.join(self.cpuset_path, "cpu" + str(cpu_id))
            for cpu_id in self._cpu_id_list()]

  def _cpu_id_list(self):
    """Extract IDs of available CPUs and return them as a list.

    The first one will be always used for all non-exclusive processes.
    :return: list[int]
    """
    cpu_list = []  # types: list[int]
    with open(self.cpuset_path + "cpuset.cpus", "rt") as cpu_def:
      for cpu_def_split in cpu_def.read().strip().split(","):
        # IDs can be in form "0-4" or "0,1,2,3,4"
        if "-" in cpu_def_split:
          a, b = map(int, cpu_def_split.split("-"))
          cpu_list.extend(range(a, b + 1)) # because cgroup's range is inclusive
          continue
        cpu_list.append(int(cpu_def_split))
    return cpu_list

  def _move_to_exclusive_cpu(self, pid):
    """Try all exclusive CPUs and place the ``pid`` to the first available one.

    :return: int, cpu_id of used CPU, -1 if placement was not possible
    """
    exclusive_cpu_list = self._cpu_id_list()[1:]
    for exclusive_cpu in exclusive_cpu_list:
      # gather tasks assigned to current exclusive CPU
      task_path = os.path.join(self.cpuset_path, "cpu" + str(exclusive_cpu), "tasks")
      with open(task_path, "rt") as fi:
        task_list = fi.read().split()
      if len(task_list) > 0:
        continue  # skip occupied CPUs
      return self._move_task(pid, exclusive_cpu)[1]
    return -1

  def _move_task(self, pid, cpu_id, cpu_mode="performance"):
    """Move ``pid`` to ``cpu_id``.

    cpu_mode can be "performance" or "powersave"
    """
    known_cpu_mode_list = ("performance", "powersave")
    with open(os.path.join(self.cpuset_path, "cpu" + str(cpu_id), "tasks"), self.task_write_mode) as fo:
      fo.write(str(pid) + "\n")
    # set the core to `cpu_mode`
    scaling_governor_file = "/sys/devices/system/cpu/cpu{:d}/cpufreq/scaling_governor".format(cpu_id)
    if os.path.exists(scaling_governor_file):
      if cpu_mode not in known_cpu_mode_list:
        logger.warning("Cannot set CPU to mode \"{}\"! Known modes {!s}".format(
          cpu_mode, known_cpu_mode_list))
      else:
        try:
          with open(scaling_governor_file, self.task_write_mode) as fo:
            fo.write(cpu_mode + "\n")  # default is "powersave"
        except IOError as e:
          # handle permission error
          logger.error("Failed to write \"{}\" to {}".format(cpu_mode, scaling_governor_file))
    return pid, cpu_id

  def _prepare_folder(self, folder):
    """If-Create folder and set group write permission."""
    if not os.path.exists(folder):
      os.mkdir(folder)
      # make your life and testing easier and create mandatory files if they don't exist
      mandatory_file_list = ("tasks", "cpuset.cpus")
      for mandatory_file in mandatory_file_list:
        file_path = os.path.join(folder, mandatory_file)
        if not os.path.exists(file_path):
          with open(file_path, "wb"):
            pass  # touche
    return folder


def read_file(path, mode="rt"):
  with open(path, mode) as fi:
    return fi.read()


def write_file(content, path, mode="wt"):
  with open(path, mode) as fo:
    fo.write(content)


import time

import slapos
from slapos.slap.slap import ComputerPartition, SoftwareInstance
from slapos.util import loads, dumps
from slapos.util import bytes2str, \
    xml2dict, dict2xml

import six
from six.moves.urllib.parse import urlparse

from flask import g, current_app, request, url_for
from slapos.proxy.db_version import DB_VERSION
from slapos.slap.slap import DEFAULT_SOFTWARE_TYPE, OLD_DEFAULT_SOFTWARE_TYPE

class NotFoundPartitionFailure(Exception):
  pass

class PartitionDeletionFailure(Exception):
  pass

class AllocationFailure(Exception):
  pass

class ConfigurationError(Exception):
  pass


def execute_db(table, query, args=(), one=False, db_version=DB_VERSION, db=None):
  if not db:
    db = g.db
  query = query % (table + db_version,)
  # current_app.logger.debug(query)
  # try:
  cur = db.execute(query, args)
  # except Exception:
  #   current_app.logger.error(
  #     'There was some issue during processing query %r on table %r with args %r',
  #     query, table, args)
  #   raise
  rv = ({cur.description[idx][0]: value
    for idx, value in enumerate(row)} for row in cur)
  return next(rv, None) if one else list(rv)

def checkIfMasterIsCurrentMaster(master_url):
  """
  Because there are several ways to contact this server, we can't easily check
  in a request() if master_url is ourself or not. So we contact master_url,
  and if it returns an ID we know: it is ourself
  """
  # Dumb way: compare with listening host/port
  host = request.host
  port = request.environ['SERVER_PORT']
  if master_url == 'http://%s:%s/' % (host, port):
    return True

  # Hack way: call ourself
  slap = slapos.slap.slap()
  slap.initializeConnection(master_url)
  try:
    return current_app.config['run_id'] == bytes2str(slap._connection_helper.GET('/getRunId'))
  except Exception:
    return False


def formatFromDB(computer_reference, partition_list,
                 computer_address=None,
                 computer_netmask=None):
  execute_db('computer', 'INSERT OR REPLACE INTO %s values(?, ?, ?)',
             (computer_reference, computer_address, computer_netmask))

  # remove references to old partitions.
  execute_db(
    'partition',
    'DELETE FROM %s WHERE computer_reference = ? and reference not in ({})'.format(
      ','.join('?' * len(partition_list)) # Create as many placeholder as partitions requested
    ),
    # Prepare arguments : first is for computer_reference, followed by the same of the partitions
    [computer_reference] + [x['partition_id'] for x in partition_list]
  )

  execute_db(
    'slave',
    'DELETE FROM %s WHERE computer_reference = ? and hosted_by not in ({})'.format(
      ','.join('?' * len(partition_list)) # Create as many placeholder as partitions requested
    ),
    # Prepare arguments : first is for computer_reference, followed by the same of the partitions
    [computer_reference] + [x['partition_id'] for x in partition_list]
  )
  execute_db('partition_network', 'DELETE FROM %s WHERE computer_reference = ?', (computer_reference,))

  for partition in partition_list:
    partition['computer_reference'] = computer_reference
    execute_db('partition', 'INSERT OR IGNORE INTO %s (reference, computer_reference) values(:partition_id, :computer_reference)', partition)
    for ip in partition['ip_list']:
      execute_db(
        'partition_network',
        'INSERT OR REPLACE INTO %s (reference, partition_reference, computer_reference, address, netmask) values(?, ?, ?, ?, ?)',
        (ip['network-interface'], partition['partition_id'], computer_reference,
         ip['ip-address'], ip['netmask'])
      )


def supplyFromDB(computer_reference, software_release_url, state):
  if state not in ('available', 'destroyed'):
    raise ValueError("Wrong state %s" % state)

  execute_db(
    'software',
    'INSERT OR REPLACE INTO %s VALUES(?, ?, ?)',
    [software_release_url, computer_reference, state])


def removeFromDB(computer_reference, software_release_url):
  execute_db(
    'software',
    'DELETE FROM %s WHERE url = ? and computer_reference=? ',
    [software_release_url, computer_reference])


def freePartitionFromDB(computer_partition_id, computer_id):
  partition = getPartitionFromDB(computer_partition_id, computer_id)
  if not partition:
    raise NotFoundPartitionFailure("Unknown partition %r on %r" % (computer_partition_id, computer_id))

  # Implement something similar to Alarm_garbageCollectDestroyUnlinkedInstance, if root instance
  # is destroyed, we request child instances in deleted state
  if not partition['requested_by']:
    args = partition['partition_reference'],
    child_partitions = execute_db(
      'partition',
      'SELECT partition_reference'
      ' FROM %s WHERE requested_by=?'
      ' ORDER BY partition_reference',
      args)
    if child_partitions:
      execute_db(
        'partition',
        "UPDATE %s SET requested_state='destroyed' WHERE requested_by=?",
        args)
      raise PartitionDeletionFailure("Not destroying yet because this partition has child partitions: "
                                     + ', '.join(p['partition_reference'] for p in child_partitions))

  execute_db(
    "partition",
    "UPDATE %s SET"
      " slap_state='free',"
      " software_release=NULL,"
      " xml=NULL,"
      " connection_xml=NULL,"
      " slave_instance_list=NULL,"
      " software_type=NULL,"
      " partition_reference=NULL,"
      " requested_by='',"
      " requested_state='started'"
    " WHERE reference=? AND computer_reference=?",
    (computer_partition_id, computer_id))

def getPartitionFromDB(reference, computer_reference):
  partition = execute_db('partition',
    'SELECT * FROM %s WHERE reference=? AND computer_reference=?',
    (reference, computer_reference), one=True)
  if partition:
    return partition
  current_app.logger.warning("Nonexisting partition %r on %r. Known are %s",
    reference, computer_reference,
    execute_db("partition", "select reference, requested_by from %s"))



def getAllocatedInstance(partition_reference, requested_by=''):
  """
  Look for existence of instance, if so return the
  corresponding partition dict, else return None
  """
  return execute_db('partition',
    'SELECT * FROM %s WHERE partition_reference=? AND requested_by=?',
    (partition_reference, requested_by), one=True)

def getAllocatedSlaveInstance(slave_reference, requested_computer_id):
  """
  Look for existence of instance, if so return the
  corresponding partition dict, else return None

  # XXX: Scope currently depends on instance which requests slave.
  # Meaning that two different instances requesting the same slave will
  # result in two different allocated slaves.
  """
  return execute_db('slave',
    'SELECT * FROM %s WHERE reference=? AND computer_reference=?',
    (slave_reference, requested_computer_id), one=True)

def requestSlave(software_release, software_type, partition_reference, requester_id, requested_by, partition_parameter_kw, filter_kw, requested_state):
  """
  Function to organise link between slave and master.
  Slave information are stored in places:
  1. slave table having information such as slave reference,
      connection information to slave (given by slave master),
      hosted_by and asked_by reference.
  2. A dictionary in slave_instance_list of selected slave master
      in which are stored slave_reference, software_type, slave_title and
      partition_parameter_kw stored as individual keys.
  """
  requested_computer_id = filter_kw['computer_guid']
  instance_xml = dict2xml(partition_parameter_kw)

  # We will search for a master corresponding to request
  args = []
  a = args.append
  q = 'SELECT * FROM %s WHERE software_release=? and computer_reference=?'
  a(software_release)
  a(requested_computer_id)
  if software_type:
    q += ' AND software_type=?'
    a(software_type)
  if 'instance_guid' in filter_kw:
    q += ' AND reference=?'
    # instance_guid should be like: %s-%s % (requested_computer_id, partition_id)
    # But code is convoluted here, so we check
    instance_guid = filter_kw['instance_guid']
    if instance_guid.startswith(requested_computer_id):
      a(instance_guid[len(requested_computer_id) + 1:])
    else:
      a(instance_guid)

  partition = execute_db('partition', q, args, one=True)
  if partition is None:
    current_app.logger.warning('No partition corresponding to slave request: %s' % args)
    raise AllocationFailure('No partition corresponding to slave request: %s' % args)

  # We set slave dictionary as described in docstring
  new_slave = {}
  slave_reference = requested_by + '_' + partition_reference
  new_slave['slave_title'] = slave_reference
  new_slave['slap_software_type'] = software_type
  new_slave['slave_reference'] = slave_reference

  for key in partition_parameter_kw:
    new_slave[key] = partition_parameter_kw[key]

  # Add slave to partition slave_list if not present else replace information
  slave_updated_or_added = False
  slave_instance_list = partition['slave_instance_list']
  if requested_state == 'destroyed':
    if slave_instance_list:
      slave_instance_list = loads(slave_instance_list.encode('utf-8'))
    before_count = len(slave_instance_list)
    slave_instance_list = [x for x in slave_instance_list if x['slave_reference'] != slave_reference]
    if before_count != len(slave_instance_list):
      slave_updated_or_added = True
  else:
    if slave_instance_list:
      slave_instance_list = loads(slave_instance_list.encode('utf-8'))
      for i, x in enumerate(slave_instance_list):
        if x['slave_reference'] == slave_reference:
          if slave_instance_list[i] != new_slave:
            slave_instance_list[i] = new_slave
            slave_updated_or_added = True
          break
      else:
        slave_instance_list.append(new_slave)
        slave_updated_or_added = True
    else:
      slave_instance_list = [new_slave]
      slave_updated_or_added = True

  q += ' WHERE reference=? AND computer_reference=?'
  a(partition['reference'])
  a(partition['computer_reference'])
  # Update slave_instance_list in database
  args = []
  a = args.append
  q = 'UPDATE %s SET slave_instance_list=?'
  a(bytes2str(dumps(slave_instance_list)))
  if slave_updated_or_added:
    timestamp = time.time()
    q += ', timestamp=?'
    a(timestamp)
  q += ' WHERE reference=? and computer_reference=?'
  a(partition['reference'])
  a(requested_computer_id)
  execute_db('partition', q, args)
  partition = getPartitionFromDB(partition['reference'], requested_computer_id)

  # Add slave to slave table if not there
  slave = execute_db('slave', 'SELECT * FROM %s WHERE reference=? and computer_reference=?',
                     [slave_reference, requested_computer_id], one=True)
  if slave is None:
    execute_db('slave',
               'INSERT INTO %s (reference,computer_reference,asked_by,hosted_by) values(?,?,?,?)',
               (slave_reference, requested_computer_id, requested_by, partition['reference']))
    slave = execute_db('slave', 'SELECT * FROM %s WHERE reference=? and computer_reference=?',
                       [slave_reference, requested_computer_id], one=True)

  address_list = []
  for address in execute_db('partition_network',
                            'SELECT * FROM %s WHERE partition_reference=? and computer_reference=?',
                            [partition['reference'], partition['computer_reference']]):
    address_list.append((address['reference'], address['address']))

  # XXX it should be ComputerPartition, not a SoftwareInstance
  return SoftwareInstance(
    _connection_dict=xml2dict(slave['connection_xml']),
    _parameter_dict=xml2dict(instance_xml),
    slap_computer_id=partition['computer_reference'],
    slap_computer_partition_id=slave['hosted_by'],
    slap_software_release_url=partition['software_release'],
    slap_server_url='slap_server_url',
    slap_software_type=partition['software_type'],
    ip_list=address_list)

def requestNotSlave(software_release, software_type, partition_reference,
                    requester_id, requested_by, partition_parameter_kw, filter_kw,
                    requested_state):
  instance_xml = dict2xml(partition_parameter_kw)
  requested_computer_id = filter_kw['computer_guid']

  partition = getAllocatedInstance(partition_reference, requested_by)
  args = []
  a = args.append
  q = 'UPDATE %s SET slap_state="busy"'

  if partition is None:
    partition = execute_db('partition',
        'SELECT * FROM %s WHERE slap_state="free" and computer_reference=?',
        [requested_computer_id], one=True)
    if partition is None:
      current_app.logger.warning('No more free computer partition')
      raise AllocationFailure('No free computer partition found for: %s' % partition_reference)
    if partition_reference:
      q += ' ,partition_reference=?'
      a(partition_reference)
    if requested_by:
      q += ' ,requested_by=?'
      a(requested_by)
    if not software_type:
      software_type = DEFAULT_SOFTWARE_TYPE
  else:
    if requested_by:
      root_partition = getAllocatedInstance(requested_by)
      if root_partition and root_partition['requested_state'] != "started":
        # propagate parent state to child
        # child can be stopped or destroyed while parent is started
        requested_state = root_partition['requested_state']

  timestamp = partition['timestamp']
  changed = timestamp is None
  for k, v in (('requested_state', requested_state or
                                   partition['requested_state']),
               ('software_release', software_release),
               ('software_type', software_type),
               ('xml', instance_xml)):
    if partition[k] != v:
      q += ', %s=?' % k
      a(v)
      changed = True

  if changed:
    timestamp = time.time()
    q += ', timestamp=?'
    a(timestamp)

  q += ' WHERE reference=? AND computer_reference=?'
  a(partition['reference'])
  a(partition['computer_reference'])

  execute_db('partition', q, args)
  partition = getPartitionFromDB(partition['reference'],
                                 partition['computer_reference'])
  address_list = []
  for address in execute_db('partition_network', 'SELECT * FROM %s WHERE partition_reference=?', [partition['reference']]):
    address_list.append((address['reference'], address['address']))

  # XXX it should be ComputerPartition, not a SoftwareInstance
  parameter_dict = xml2dict(partition['xml'])
  parameter_dict['timestamp'] = str(partition['timestamp'])
  return SoftwareInstance(
    _connection_dict=xml2dict(partition['connection_xml']),
    _parameter_dict=parameter_dict,
    connection_xml=partition['connection_xml'],
    slap_computer_id=partition['computer_reference'],
    slap_computer_partition_id=partition['reference'],
    slap_software_release_url=partition['software_release'],
    slap_server_url='slap_server_url',
    slap_software_type=partition['software_type'],
    _instance_guid='%(computer_reference)s-%(reference)s' % partition,
    _requested_state=requested_state or 'started',
    ip_list=address_list)

def isRequestToBeForwardedToExternalMaster(parsed_request_dict):
    """
    Check if we HAVE TO forward the request.
    Several cases:
     * The request specifies a master_url in filter_kw
     * The software_release of the request is in a automatic forward list
    """
    master_url = parsed_request_dict['filter_kw'].get('master_url')

    if checkMasterUrl(master_url):
      # Don't allocate the instance locally, but forward to specified master
      return master_url

    software_release = parsed_request_dict['software_release']
    for mutimaster_url, mutimaster_entry in six.iteritems(current_app.config.get('multimaster', {})):
      if software_release in mutimaster_entry['software_release_list']:
        # Don't allocate the instance locally, but forward to specified master
        return mutimaster_url
    return None

def forwardRequestToExternalMaster(master_url, parsed_request_dict):
  """
  Forward instance request to external SlapOS Master.
  """
  master_entry = current_app.config.get('multimaster').get(master_url, {})
  key_file = master_entry.get('key')
  cert_file = master_entry.get('cert')
  if master_url.startswith('https') and (not key_file or not cert_file):
    current_app.logger.warning('External master %s configuration did not specify key or certificate.' % master_url)
    raise ConfigurationError('External master %s configuration did not specify key or certificate.' % master_url)
  if master_url.startswith('https') and not master_url.startswith('https') and (key_file or cert_file):
    current_app.logger.warning('External master %s configuration specifies key or certificate but is using plain http.' % master_url)
    raise ConfigurationError('External master %s configuration specifies key or certificate but is using plain http.' % master_url)

  slap = slapos.slap.slap()
  if key_file:
    slap.initializeConnection(master_url, key_file=key_file, cert_file=cert_file)
  else:
    slap.initializeConnection(master_url)

  partition_reference = parsed_request_dict['partition_reference']

  filter_kw = parsed_request_dict['filter_kw']
  partition_parameter_kw = parsed_request_dict['partition_parameter_kw']

  state = parsed_request_dict['requested_state']
  current_app.logger.info("Forwarding request of %s (state=%s) to %s ", partition_reference, state, master_url)
  current_app.logger.debug("parsed_request_dict: %s", parsed_request_dict)

  if master_entry.get('computer') and master_entry.get('partition'):
    current_app.logger.debug("requesting from partition %s", master_entry)
    # XXX ComputerPartition.request and OpenOrder.request have different signatures
    partition = slap.registerComputerPartition(
        master_entry['computer'],
        master_entry['partition'],
    ).request(
        software_release=parsed_request_dict['software_release'],
        software_type=parsed_request_dict['software_type'],
        partition_reference=partition_reference,
        shared=parsed_request_dict['slave'],
        partition_parameter_kw=partition_parameter_kw,
        filter_kw=filter_kw,
        state=state,
    )
  else:
    filter_kw['source_instance_id'] = partition_reference
    partition = slap.registerOpenOrder().request(
        software_release=parsed_request_dict['software_release'],
        partition_reference=partition_reference,
        partition_parameter_kw=partition_parameter_kw,
        software_type=parsed_request_dict['software_type'],
        filter_kw=filter_kw,
        state=state,
        shared=parsed_request_dict['slave'],
    )

  # Store in database
  if state == 'destroyed':
    execute_db(
      'forwarded_partition_request',
      'DELETE FROM %s WHERE partition_reference = :partition_reference and master_url = :master_url',
      {'partition_reference':partition_reference, 'master_url': master_url})
  else:
    execute_db(
      'forwarded_partition_request',
      'INSERT OR REPLACE INTO %s values(:partition_reference, :master_url)',
      {'partition_reference':partition_reference, 'master_url': master_url})

  # XXX move to other end
  partition._master_url = master_url # type: ignore
  partition._connection_helper = None
  partition._software_release_document = parsed_request_dict['software_release'] # type: ignore

  return partition

def checkMasterUrl(master_url):
  """
  Check if master_url doesn't represent ourself, and check if it is whitelisted
  in multimaster configuration.
  """
  if not master_url:
    return False

  if checkIfMasterIsCurrentMaster(master_url):
    # master_url is current server: don't forward
    return False

  master_entry = current_app.config.get('multimaster').get(master_url, None)
  # Check if this master is known
  if not master_entry:
    # Check if it is ourself
    if not master_url.startswith('https') and checkIfMasterIsCurrentMaster(master_url):
      return False
    current_app.logger.warning('External SlapOS Master URL %s is not listed in multimaster list.' % master_url)
    raise ConfigurationError('External SlapOS Master URL %s is not listed in multimaster list.' % master_url)

  return True

def requestInstanceFromDB(slave=None, **parsed_request_dict):
  # Check first if instance is already allocated
  if slave:
    # slapproxy cannot request frontends, but we can workaround common cases,
    # so that during tests promises are succesful.
    if not isRequestToBeForwardedToExternalMaster(parsed_request_dict):
      # if client request a "simple" frontend for an URL, we can tell this
      # client to use the URL directly.
      apache_frontend_sr_url_list = (
          'http://git.erp5.org/gitweb/slapos.git/blob_plain/HEAD:/software/apache-frontend/software.cfg',
      )
      if parsed_request_dict['software_release'] in apache_frontend_sr_url_list \
        and parsed_request_dict.get('software_type', '') in ('', OLD_DEFAULT_SOFTWARE_TYPE, DEFAULT_SOFTWARE_TYPE):
        url_parameter = parsed_request_dict['partition_parameter_kw'].get('url')
        if url_parameter:
          if request.scheme == 'https':
            # Only handle the secure access if slapproxy is also
            # accessed with secure https
            # to ensure not lowering the connection
            parsed_url_parameter = urlparse(url_parameter)
            # XXX hardcoded http_proxy. set in views.py
            parsed_secure_access_url = urlparse(url_for(
              'httpproxy.proxy_request',
              url_scheme=parsed_url_parameter.scheme,
              url_netloc=parsed_url_parameter.netloc,
              url_path=parsed_url_parameter.path,
              _external=True
            ))
            secure_access_url = parsed_secure_access_url._replace(
              query=parsed_url_parameter.query,
              fragment=parsed_url_parameter.fragment
            ).geturl()
          else:
            # If slaproxy is not accessed with https
            # return the original url
            secure_access_url = url_parameter
            parsed_secure_access_url = urlparse(secure_access_url)
          current_app.logger.warning("Bypassing frontend for %s => %s", parsed_request_dict, url_parameter)
          partition = ComputerPartition('', 'Fake frontend for {}'.format(url_parameter))
          partition.slap_computer_id = ''
          partition.slap_computer_partition_id = ''
          partition._parameter_dict = {}
          partition._connection_dict = {
            'secure_access': secure_access_url,
            'domain': parsed_secure_access_url.netloc,
          }
          return partition
      # another similar case is for KVM frontends. This is used in
      # request-slave-frontend from software/kvm/instance-kvm.cfg.jinja2
      # requested values by 'return' recipe are: url resource port domainname
      kvm_frontend_sr_url_list = (
          'http://git.erp5.org/gitweb/slapos.git/blob_plain/refs/tags/slapos-0.92:/software/kvm/software.cfg',
      )
      if parsed_request_dict['software_release'] in kvm_frontend_sr_url_list \
          and parsed_request_dict.get('software_type') in ('frontend', ):
        host = parsed_request_dict['partition_parameter_kw'].get('host')
        port = parsed_request_dict['partition_parameter_kw'].get('port')
        if host and port:
          # host is supposed to be ipv6 without brackets.
          if ':' in host and host[0] != '[':
            host = '[%s]' % host
          url = 'https://%s:%s/' % (host, port)
          current_app.logger.warning("Bypassing KVM VNC frontend for %s => %s", parsed_request_dict, url)
          partition = ComputerPartition('', 'Fake KVM VNC frontend for {}'.format(url))
          partition.slap_computer_id = ''
          partition.slap_computer_partition_id = ''
          partition._parameter_dict = {}
          partition._connection_dict = {
            'url': url,
            'domainname': host,
            'port': port,
            'path': '/'
          }
          return partition

    # XXX: concat is unreliable
    slave_reference = parsed_request_dict['requested_by'] + '_' \
                    + parsed_request_dict['partition_reference']
    requested_computer_id = parsed_request_dict['filter_kw'].get('computer_guid', current_app.config['computer_id'])
    matching_partition = getAllocatedSlaveInstance(slave_reference, requested_computer_id)
  else:
    matching_partition = getAllocatedInstance(
      parsed_request_dict['partition_reference'],
      parsed_request_dict['requested_by'])

  if not matching_partition:
    # Instance is not yet allocated: try to do it.
    external_master_url = isRequestToBeForwardedToExternalMaster(parsed_request_dict)
    if external_master_url:
      return forwardRequestToExternalMaster(external_master_url, {
        # Prefix instance reference with id of requester (partition id (ends with a digit) or 'user' (cannot be a partition id))
        'requested_by': parsed_request_dict['requested_by'],
        'partition_reference': '%s_%s' % (parsed_request_dict['requester_id'], parsed_request_dict['partition_reference']),
        'software_release': parsed_request_dict['software_release'],
        'software_type': parsed_request_dict['software_type'],
        'partition_parameter_kw': parsed_request_dict['partition_parameter_kw'],
        'filter_kw': parsed_request_dict['filter_kw'],
        # Note: currently ignored for slave instance (slave instances
        # are always started).
        'requested_state': parsed_request_dict['requested_state'],
        # Is it a slave instance?
        'slave': slave
      })
    # XXX add support for automatic deployment on specific node depending on available SR and partitions on each Node.
    # Note: It only deploys on default node if SLA not specified
    # XXX: split request and request slave into different update/allocate functions and simplify.

  # By default, ALWAYS request instance on default computer
  parsed_request_dict['filter_kw'].setdefault('computer_guid', current_app.config['computer_id'])
  if slave:
    software_instance = requestSlave(**parsed_request_dict)
  else:
    software_instance = requestNotSlave(**parsed_request_dict)
  return software_instance


def bangInstanceFromDB(partition_reference, requested_by):
  execute_db('partition',
    "UPDATE %s SET timestamp=?"
    " WHERE (partition_reference=? AND requested_by=?) OR requested_by=?",
    (time.time(), partition_reference, requested_by, partition_reference))


def getRootPartitionList(title=None):
  query = "SELECT * FROM %s WHERE slap_state!='free' AND requested_by=''"
  args = []
  if title:
    query += ' AND partition_reference=?'
    args.append(title)
  return execute_db('partition', query, args)

def getRootSharedList(title=None):
  query = "SELECT * FROM %s WHERE asked_by=''"
  args = []
  if title:
    query += ' AND reference=?'
    args.append('_' + title)
  return execute_db('slave', query, args)

def getInstanceTreeList(title=None):
  return [getRootPartitionList(), getRootSharedList()]



from flask import Blueprint, abort, send_from_directory, current_app

panel_blueprint = Blueprint('panel', __name__, static_url_path='', static_folder='static_panel')

@panel_blueprint.route('/', methods=['GET'])
def index():
  return panel_blueprint.send_static_file('panel.html')

@panel_blueprint.route('/public/<path:filename>')
def download_public_file(filename):
  config_parameter_key = 'PUBLIC_DIRECTORY_PATH'
  if config_parameter_key in current_app.config:
    return send_from_directory(
      current_app.config[config_parameter_key],
      filename,
      # As the content of the public directory is unknown,
      # prevent publishing any kind of website from the proxy
      as_attachment=True,
      # handle 304 status
      conditional=True
    )
  else:
    # 404 is not used, to help understanding that the
    # panel is not configured to search the public directory
    return abort(403, 'The public directory path is not configured.')



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2010-2014 Vifib SARL and Contributors.
# All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

import sys
import six

from slapos.cli.command import resetLogger
from slapos.cli.config import ClientConfigCommand
from slapos.client import init, ClientConfig


class ListCommand(ClientConfigCommand):
    """List all available compute nodes"""

    def get_parser(self, prog_name):
        ap = super(ListCommand, self).get_parser(prog_name)
        return ap

    def take_action(self, args):
        configp = self.fetch_config(args)
        conf = ClientConfig(args, configp)

        local = init(conf, self.app.log)
        do_list(self.app.log, conf, local)


def do_list(logger, conf, local):
    with resetLogger(logger):
      computer_dict = local['slap'].getComputerDict()
      if computer_dict == {}:
        logger.info('No existing computer.')
        return
      logger.info('List of Computers:')
      for title, computer in six.iteritems(computer_dict):
        logger.info('%s %s', computer._reference, title)



from flask import request, abort, Blueprint, url_for, redirect
from .db import execute_db, getRootPartitionList, getRootSharedList
from six.moves.urllib.parse import unquote, urljoin
from slapos.util import loads, dict2xml
import re

hateoas_blueprint = Blueprint('hateoas', __name__)


# hateoas routing
# ---------------

# We only need to handle the hateoas requests made by
#   slapos service list
#   slapos service info <reference>
#   slapos computer list
#   slapos computer info <reference>


def unquoted_url_for(method, **kwargs):
  return unquote(url_for(method, **kwargs))

def busy_root_partitions_list(title=None):
  partitions = []
  for row in getRootPartitionList(title=title):
    p = dict(row)
    p['url_string'] = p['software_release']
    p['title'] = p['partition_reference']
    p['relative_url'] = url_for('.hateoas_partitions', partition_reference=p['partition_reference'])
    partitions.append(p)
  return partitions

def busy_root_shared_list(title=None):
  shared = []
  for row in getRootSharedList(title=title):
    host = execute_db('partition', 'SELECT * FROM %s WHERE reference=?', [row['hosted_by']], one=True)
    if not host:
      continue
    for slave_dict in loads(host['slave_instance_list'].encode('utf-8')):
      if slave_dict['slave_reference'] == row['reference']:
        break
    else:
      continue
    s = {}
    s['url_string'] = host['software_release']
    s['title'] = row['reference'][1:] # root shared are prefixed with _
    s['relative_url'] = url_for('.hateoas_shared', shared_reference=s['title'])
    shared.append(s)
  return shared

def computers_list(reference=None):
  computers = []
  query = 'SELECT * FROM %s'
  args = []
  if reference:
    query += ' WHERE reference=?'
    args.append(reference)
  for row in execute_db('computer', query, args):
    c = dict(row)
    c['title'] = c['reference']
    c['relative_url'] = url_for('.hateoas_computers', computer_reference=c['reference'])
    computers.append(c)
  return computers

r_string = re.compile('"((\\.|[^\\"])*)"')

def is_valid(name):
  match = r_string.match(name)
  if match.group(0) == name:
    return True
  return False

p_service_list = 'portal_type:"Instance Tree" AND validation_state:validated'
p_service_info = p_service_list + ' AND title:='
p_computer_list = 'portal_type:"Compute Node" AND validation_state:validated'
p_computer_info = p_computer_list + ' AND reference:='

def parse_query(query):
  if query == p_service_list:
    return busy_root_partitions_list() + busy_root_shared_list()
  elif query.startswith(p_service_info):
    title = query[len(p_service_info):]
    if is_valid(title):
      partition = busy_root_partitions_list(title.strip('"'))
      if partition:
        return partition
      return busy_root_shared_list(title.strip('"'))
  elif query == p_computer_list:
    return computers_list()
  elif query.startswith(p_computer_info):
    reference = query[len(p_computer_info):]
    if is_valid(reference):
      return computers_list(reference.strip('"'))
  return None

@hateoas_blueprint.route('/partitions/<partition_reference>', methods=['GET'])
def hateoas_partitions(partition_reference):
  partition = execute_db('partition', 'SELECT * FROM %s WHERE partition_reference=?', [partition_reference], one=True)
  if partition is None:
    abort(404)
  partition['reference'] = partition['partition_reference']
  partition['shared'] = 0
  return hateoas_service_document(**partition)

@hateoas_blueprint.route('/shared/<shared_reference>', methods=['GET'])
def hateoas_shared(shared_reference):
  slave_reference = '_' + shared_reference # root shared are prefixed with _ in db
  shared = execute_db('slave', 'SELECT * FROM %s WHERE reference=?', [slave_reference], one=True)
  if shared is None:
    abort(404)
  partition = execute_db('partition', 'SELECT * FROM %s WHERE reference=?', [shared['hosted_by']], one=True)
  if partition is None:
    abort(404)
  slave_list = loads(partition['slave_instance_list'].encode('utf-8'))
  for slave_dict in slave_list:
    if slave_dict['slave_reference'] == slave_reference:
      break
  else:
    abort(404)
  del slave_dict['slave_title'], slave_dict['slave_reference']
  software_type = slave_dict.pop('slap_software_type')
  xml = dict2xml(slave_dict)
  return hateoas_service_document(
    reference = shared_reference,
    requested_state='unused',
    xml=xml,
    connection_xml=shared['connection_xml'],
    software_release=partition['software_release'],
    software_type=software_type,
    shared=1,
  )

def hateoas_service_document(**kw):
  # my_slap_state corresponds to requested_state, not slap_state.
  return {
    '_embedded': {
      '_view': {
        'form_id': {
          'type': 'StringField',
          'key': 'form_id',
          'default': 'InstanceTree_viewAsHateoas',
        },
        'my_reference': {
          'type': 'StringField',
          'key': 'field_my_reference',
          'default': kw['reference'],
        },
        'my_slap_state': {
          'type': 'StringField',
          'key': 'field_my_slap_state',
          'default': kw['requested_state'],
        },
        'my_text_content': {
          'type': 'StringField',
          'key': 'field_my_text_content',
          'default': kw['xml'],
        },
        'my_connection_parameter_list': {
          'type': 'StringField',
          'key': 'field_my_connection_parameter_list',
          'default': kw['connection_xml'],
        },
        'my_url_string': {
          'type': 'StringField',
          'key': 'field_my_url_string',
          'default': kw['software_release'],
        },
        'my_source_reference': {
          'type': 'StringField',
          'key': 'field_my_source_reference',
          'default': kw['software_type'],
        },
        'my_root_slave': {
          'type': 'IntegerField',
          'key': 'field_my_root_slave',
          'default': kw['shared'],
        },
      },
    },
    '_links': {
      'type': {
        'name': 'Instance Tree',
      },
    },
  }

@hateoas_blueprint.route('/computers/<computer_reference>', methods=['GET'])
def hateoas_computers(computer_reference):
  computer = execute_db('computer', 'SELECT * FROM %s WHERE reference=?', [computer_reference], one=True)
  if computer is None:
    abort(404)
  return {
    '_embedded': {
      '_view': {
        'form_id': {
          'type': 'StringField',
          'key': 'computer',
          'default': computer['reference'],
        },
        'my_reference': {
          'type': 'StringField',
          'key': 'reference',
          'default': computer['reference'],
        },
        'my_title': {
          'type': 'StringField',
          'key': 'reference',
          'default': computer['reference'],
        },
      },
    },
    '_links': {
      'type': {
        'name': 'Computer',
      },
    },
  }

def hateoas_traverse():
  return redirect(request.args['relative_url'])

def hateoas_search():
  contents = parse_query(request.args["query"])
  if contents is None:
    abort(400, "Unhandled query: %s" % request.args["query"])
  return { '_embedded': {'contents': contents} }

def hateoas_root():
  return {
    '_links': {
      'raw_search': {
      'href': urljoin(request.host_url, unquoted_url_for('.hateoas', mode='search', query='{query}', select_list='{select_list}'))
    },
      'traverse': {
        'href': urljoin(request.host_url, unquoted_url_for('.hateoas', mode='traverse', relative_url='{relative_url}', view='{view}'))
      },
    }
  }

mode_handlers = {
  None: hateoas_root,
  'search': hateoas_search,
  'traverse': hateoas_traverse,
}

@hateoas_blueprint.route('/', methods=['GET'])
def hateoas():
  mode = request.args.get('mode')
  handler = mode_handlers.get(mode, lambda: abort(400, "Unhandled mode: %s" % mode))
  resp = handler()
  return resp



from flask import request, Blueprint, current_app, abort
from .db import execute_db, requestInstanceFromDB, supplyFromDB, removeFromDB, \
                bangInstanceFromDB, freePartitionFromDB, formatFromDB, \
                NotFoundPartitionFailure, PartitionDeletionFailure, \
                getPartitionFromDB, AllocationFailure, ConfigurationError
from slapos.util import loads, dumps
from slapos.util import unicode2str, \
    xml2dict, dict2xml

from slapos.slap.slap import Computer, ComputerPartition, \
    SoftwareRelease, NotFoundError

slap_tool_blueprint = Blueprint('slap_tool', __name__)

EMPTY_DICT_XML = dumps({})

class UnauthorizedError(Exception):
  pass

def partitiondict2partition(partition):
  computer_id = partition['computer_reference']
  partition_id = partition['reference']
  slap_partition = ComputerPartition(computer_id, partition_id)
  slap_partition._software_release_document = None
  slap_partition._requested_state = 'destroyed'
  slap_partition._need_modification = 0
  slap_partition._instance_guid = '%(computer_reference)s-%(reference)s' \
    % partition

  if partition['software_release']:
    slap_partition._need_modification = 1
    slap_partition._requested_state = partition['requested_state']
    slap_partition._parameter_dict = xml2dict(partition['xml'])
    address_list = []
    full_address_list = []
    for address in execute_db('partition_network',
                              'SELECT * FROM %s WHERE partition_reference=? AND computer_reference=?',
                              (partition_id, computer_id)):
      address_list.append((address['reference'], address['address']))
    slap_partition._parameter_dict['ip_list'] = address_list
    slap_partition._parameter_dict['full_address_list'] = full_address_list
    slap_partition._parameter_dict['slap_software_type'] = \
        partition['software_type']
    slap_partition._parameter_dict['instance_title'] = \
        partition['partition_reference']
    slap_partition._parameter_dict['root_instance_title'] = \
        partition['requested_by'] or partition['partition_reference']
    slap_partition._parameter_dict['slap_computer_id'] = \
        computer_id
    slap_partition._parameter_dict['slap_computer_partition_id'] = \
        partition_id
    slap_partition._parameter_dict['slap_software_release_url'] = \
        partition['software_release']

    if partition['slave_instance_list'] is not None:
      slap_partition._parameter_dict['slave_instance_list'] = \
          loads(partition['slave_instance_list'].encode('utf-8'))
    else:
      slap_partition._parameter_dict['slave_instance_list'] = []
    timestamp = partition['timestamp']
    if timestamp:
      slap_partition._parameter_dict['timestamp'] = str(timestamp)
    slap_partition._connection_dict = xml2dict(partition['connection_xml'])
    slap_partition._software_release_document = SoftwareRelease(
      software_release=partition['software_release'],
      computer_guid=computer_id)

  return slap_partition

@slap_tool_blueprint.route('/getComputerInformation', methods=['GET'])
def getComputerInformation():
  # Kept only for backward compatiblity
  return getFullComputerInformation()


@slap_tool_blueprint.route('/getFullComputerInformation', methods=['GET'])
def getFullComputerInformation():
  computer_id = request.args['computer_id']
  computer_list = execute_db('computer', 'SELECT * FROM %s WHERE reference=?', [computer_id])
  if len(computer_list) != 1:
    # Backward compatibility
    if computer_id != current_app.config['computer_id']:
      raise NotFoundError('%s is not registered.' % computer_id)
  slap_computer = Computer(computer_id)
  slap_computer._software_release_list = []
  for sr in execute_db('software', 'select * from %s WHERE computer_reference=?', [computer_id]):
    software_release = SoftwareRelease(
        software_release=sr['url'],
        computer_guid=computer_id)
    software_release._requested_state = sr['requested_state']
    slap_computer._software_release_list.append(software_release)
  slap_computer._computer_partition_list = []
  for partition in execute_db('partition', 'SELECT * FROM %s WHERE computer_reference=?', [computer_id]):
    slap_computer._computer_partition_list.append(partitiondict2partition(
      partition))
  return dumps(slap_computer)

@slap_tool_blueprint.route('/registerComputerPartition', methods=['GET'])
def registerComputerPartition():
  args = request.args
  partition = getPartitionFromDB(unicode2str(args['computer_partition_reference']),
                                 unicode2str(args['computer_reference']))
  if partition is None:
    raise UnauthorizedError
  return dumps(partitiondict2partition(partition))


@slap_tool_blueprint.route('/setComputerPartitionConnectionXml', methods=['POST'])
def setComputerPartitionConnectionXml():
  slave_reference = request.form.get('slave_reference', None)
  computer_partition_id = unicode2str(request.form['computer_partition_id'])
  computer_id = unicode2str(request.form['computer_id'])
  connection_xml = dict2xml(loads(request.form['connection_xml'].encode('utf-8')))
  if not slave_reference or slave_reference == 'None':
    query = 'UPDATE %s SET connection_xml=? WHERE reference=? AND computer_reference=?'
    argument_list = [connection_xml, computer_partition_id, computer_id]
    execute_db('partition', query, argument_list)
  else:
    query = 'UPDATE %s SET connection_xml=? , hosted_by=? WHERE reference=?'
    argument_list = [connection_xml, computer_partition_id, slave_reference]
    execute_db('slave', query, argument_list)
  return 'done'

@slap_tool_blueprint.route('/buildingSoftwareRelease', methods=['POST'])
def buildingSoftwareRelease():
  return 'Ignored'

@slap_tool_blueprint.route('/destroyedSoftwareRelease', methods=['POST'])
def destroyedSoftwareRelease():
  removeFromDB(request.form['computer_id'], request.form['url'])
  return 'OK'

@slap_tool_blueprint.route('/availableSoftwareRelease', methods=['POST'])
def availableSoftwareRelease():
  return 'Ignored'

@slap_tool_blueprint.route('/softwareReleaseError', methods=['POST'])
def softwareReleaseError():
  return 'Ignored'

@slap_tool_blueprint.route('/softwareInstanceError', methods=['POST'])
def softwareInstanceError():
  return 'Ignored'

@slap_tool_blueprint.route('/softwareInstanceBang', methods=['POST'])
def softwareInstanceBang():
  requested_by = getRequesterFromForm(request.form, True)
  bangInstanceFromDB(requested_by, '')
  return 'OK'

@slap_tool_blueprint.route('/startedComputerPartition', methods=['POST'])
def startedComputerPartition():
  return 'Ignored'

@slap_tool_blueprint.route('/stoppedComputerPartition', methods=['POST'])
def stoppedComputerPartition():
  return 'Ignored'

@slap_tool_blueprint.route('/destroyedComputerPartition', methods=['POST'])
def destroyedComputerPartition():
  key = (unicode2str(request.form['computer_partition_id']),
         unicode2str(request.form['computer_id']))
  try:
    freePartitionFromDB(*key)
  except (NotFoundPartitionFailure, PartitionDeletionFailure) as error:
    return str(error)

  return 'OK'

@slap_tool_blueprint.route('/useComputer', methods=['POST'])
def useComputer():
  return 'Ignored'

@slap_tool_blueprint.route('/loadComputerConfigurationFromXML', methods=['POST'])
def loadComputerConfigurationFromXML():
  xml = request.form['xml']
  computer_dict = loads(xml.encode('utf-8'))
  partition_list = []
  for input_partition in computer_dict['partition_list']:
    ip_list = []
    for input_address in input_partition['address_list']:
      ip_list.append({
        'ip-address': input_address['addr'],
        'netmask': input_address['netmask'],
        # keep "or input_partition['reference']" for backward compatibility in webrunner
        'network-interface': input_partition['tap']['name'] or input_partition['reference']
      })
    partition_list.append({
      'partition_id': input_partition['reference'],
      'ip_list': ip_list
    })
  formatFromDB(
    computer_dict['reference'],
    partition_list,
    computer_address=computer_dict['address'],
    computer_netmask=computer_dict['netmask']
  )
  return 'done'



@slap_tool_blueprint.route('/supplySupply', methods=['POST'])
def supplySupply():
  url = request.form['url']
  computer_id = request.form['computer_id']
  state = request.form['state']
  supplyFromDB(computer_id, url, state)
  return 'Supplied %r to be %s' % (url, state)


@slap_tool_blueprint.route('/requestComputerPartition', methods=['POST'])
def requestComputerPartition():
  form = request.form
  parsed_request_dict = {
    'requester_id': unicode2str(request.form.get('computer_partition_id', 'user')),
    'requested_by': getRequesterFromForm(form) or '',
    'software_release': unicode2str(form['software_release']),
    'software_type': unicode2str(form['software_type']),
    'partition_reference': unicode2str(form['partition_reference']),
    'partition_parameter_kw': loads(form.get('partition_parameter_xml', EMPTY_DICT_XML).encode('utf-8')),
    'filter_kw': loads(form.get('filter_xml', EMPTY_DICT_XML).encode('utf-8')),
    # Note: currently ignored for slave instance (slave instances
    # are always started).
    'requested_state': loads(form['state'].encode('utf-8')),
    # Is it a slave instance?
    'slave': loads(form.get('shared_xml', EMPTY_DICT_XML).encode('utf-8'))
  }
  try:
    partition = requestInstanceFromDB(**parsed_request_dict)
  except (AllocationFailure, ConfigurationError) as e:
    return abort(404, str(e))
  if getattr(partition, '_request_dict', None) is not None:
    # ResourceNotReady
    return abort(408, 'Resource not ready')
  return dumps(partition)

def getRequesterFromForm(form, required=False):
  try:
    computer_id = form['computer_id']
    partition_id = form['computer_partition_id']
  except KeyError:
    if required:
      raise
    return
  partition = getPartitionFromDB(unicode2str(partition_id), unicode2str(computer_id))
  return partition and (partition['requested_by'] or
                        partition['partition_reference'])


@slap_tool_blueprint.route('/softwareInstanceRename', methods=['POST'])
def softwareInstanceRename():
  new_name = unicode2str(request.form['new_name'])
  key = (unicode2str(request.form['computer_partition_id']),
         unicode2str(request.form['computer_id']))
  partition = getPartitionFromDB(*key)
  if not partition:
    return "Unknown partition %r on %r" % key

  if not partition['requested_by']:
    execute_db(
      'partition',
      'UPDATE %s SET requested_by=? WHERE requested_by=?',
      (new_name, partition['partition_reference']))

  execute_db(
    'partition',
    'UPDATE %s SET partition_reference=?'
    ' WHERE reference=? AND computer_reference=?',
    (new_name,) + key)
  return 'done'

@slap_tool_blueprint.route('/getComputerPartitionStatus', methods=['GET'])
def getComputerPartitionStatus():
  return dumps('Not implemented.')

@slap_tool_blueprint.route('/computerBang', methods=['POST'])
def computerBang():
  return dumps('')

@slap_tool_blueprint.route('/getComputerPartitionCertificate', methods=['GET'])
def getComputerPartitionCertificate():
  # proxy does not use partition certificate, but client calls this.
  return dumps({'certificate': '', 'key': ''})

@slap_tool_blueprint.route('/getSoftwareReleaseListFromSoftwareProduct', methods=['GET'])
def getSoftwareReleaseListFromSoftwareProduct():
  software_product_reference = request.args.get('software_product_reference')
  software_release_url = request.args.get('software_release_url')

  if software_release_url:
    assert(software_product_reference is None)
    raise NotImplementedError('software_release_url parameter is not supported yet.')
  else:
    assert(software_product_reference is not None)
    if software_product_reference in current_app.config['software_product_list']:
      software_release_url_list =\
          [current_app.config['software_product_list'][software_product_reference]]
    else:
      software_release_url_list = []
    return dumps(software_release_url_list)



##############################################################################
#
# Copyright (c) 2018 Vifib SARL and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

import json
import unittest
import mock
import os
import tempfile
import textwrap
import hashlib
import socket
import errno
import time
import glob
import subprocess
import multiprocessing
from contextlib import closing
try:
  from configparser import ConfigParser
except ImportError:
  # BBB python2
  from ConfigParser import SafeConfigParser as ConfigParser
  unittest.TestCase.assertRaisesRegex = unittest.TestCase.assertRaisesRegexp

import netaddr
import psutil

from slapos.slap.standalone import StandaloneSlapOS
from slapos.slap.standalone import SlapOSNodeSoftwareError
from slapos.slap.standalone import PartitionForwardConfiguration
from slapos.slap.standalone import PartitionForwardAsPartitionConfiguration
import slapos.util


SLAPOS_TEST_IPV4 = os.environ['SLAPOS_TEST_IPV4']
SLAPOS_TEST_IPV6 = os.environ['SLAPOS_TEST_IPV6']
SLAPOS_TEST_PORT = int(os.environ.get('SLAPOS_TEST_PORT', 33333))


def checkPortIsFree():
  """Sanity check that we did not leak a process listening on this port.
  """
  s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
  with closing(s):
    try:
      s.connect((SLAPOS_TEST_IPV4, SLAPOS_TEST_PORT))
      raise RuntimeError(
          "Port needed for tests ({SLAPOS_TEST_IPV4}:{SLAPOS_TEST_PORT}) is already in use"
          .format(**globals()))
    except socket.error as e:
      if e.errno == errno.ECONNREFUSED:
        return
      raise


class TestSlapOSStandaloneSetup(unittest.TestCase):
  def setUp(self):
    checkPortIsFree()

  def setupSimpleStandalone(self):
    working_dir = tempfile.mkdtemp(prefix=__name__)
    self.addCleanup(slapos.util.rmtree, working_dir)
    standalone = StandaloneSlapOS(
        working_dir, SLAPOS_TEST_IPV4, SLAPOS_TEST_PORT)
    self.addCleanup(standalone.stop)
    return standalone

  @staticmethod
  def getInstancePath(standalone, *segments):
    return os.path.join(standalone.instance_directory, *segments)

  def assertExists(self, path):
    self.assertTrue(os.path.exists(path))

  def assertNotExists(self, path):
    self.assertFalse(os.path.exists(path))

  @classmethod
  def getJsonResourceList(cls, standalone):
    return [
        cls.getJson(
            cls.getInstancePath(
                standalone, 'slappart%d' % i, '.slapos-resource'))
        for i in range(standalone._partition_count)]

  @staticmethod
  def getJson(path):
    with open(path) as f:
      return json.load(f)

  def test_format(self):
    standalone = self.setupSimpleStandalone()
    standalone.format(3, SLAPOS_TEST_IPV4, SLAPOS_TEST_IPV6)
    self.assertExists(standalone.software_directory)
    self.assertExists(standalone.instance_directory)
    self.assertExists(
        os.path.join(standalone.instance_directory, 'slappart0'))
    self.assertExists(
        os.path.join(standalone.instance_directory, 'slappart1'))
    self.assertExists(
        os.path.join(standalone.instance_directory, 'slappart2'))
    for i in range(3):
      self.assertExists(
          self.getInstancePath(standalone, 'slappart%d' % i, '.slapos-resource'))

  def test_format_ipv6_big_range(self):
    standalone = self.setupSimpleStandalone()
    prefixlen = 96
    slapos_fake_ipv6_range = '%s/%d' % (SLAPOS_TEST_IPV6, prefixlen)
    standalone.format(3, SLAPOS_TEST_IPV4, slapos_fake_ipv6_range)
    resource_list = self.getJsonResourceList(standalone)
    for i, resource in enumerate(resource_list):
      resource_prefixlen = int(resource['ipv6_range']['network'].split('/')[1])
      self.assertEqual(resource_prefixlen, prefixlen + 16)
      partition_ipv6 = resource['address_list'][1]['addr']
      self.assertTrue(netaddr.valid_ipv6(partition_ipv6))
      for other_resource in resource_list[i + 1:]:
        self.assertNotEqual(
            resource['ipv6_range']['addr'],
            other_resource['ipv6_range']['addr'])
        self.assertNotEqual(
            partition_ipv6,
            other_resource['address_list'][1]['addr'])

  def test_format_ipv6_small_range(self):
    standalone = self.setupSimpleStandalone()
    prefixlen = 112
    slapos_fake_ipv6_range = '%s/%d' % (SLAPOS_TEST_IPV6, prefixlen)
    addr0 = str(netaddr.IPNetwork(slapos_fake_ipv6_range).network)
    standalone.format(3, SLAPOS_TEST_IPV4, slapos_fake_ipv6_range)
    resource_list = self.getJsonResourceList(standalone)
    for i, resource in enumerate(resource_list):
      self.assertFalse(resource['ipv6_range'])
      partition_ipv6 = resource['address_list'][1]['addr']
      self.assertTrue(netaddr.valid_ipv6(partition_ipv6))
      for other_resource in resource_list[i + 1:]:
        other_partition_ipv6 = other_resource['address_list'][1]['addr']
        self.assertNotEqual(partition_ipv6, other_partition_ipv6)
        self.assertNotEqual(partition_ipv6, addr0)

  def test_format_ipv6_very_small_range(self):
    standalone = self.setupSimpleStandalone()
    prefixlen = 126
    slapos_fake_ipv6_range = '%s/%d' % (SLAPOS_TEST_IPV6, prefixlen)
    addr0 = str(netaddr.IPNetwork(slapos_fake_ipv6_range).network)
    standalone.format(8, SLAPOS_TEST_IPV4, slapos_fake_ipv6_range)
    resource_list = self.getJsonResourceList(standalone)
    for i, resource in enumerate(resource_list):
      self.assertFalse(resource['ipv6_range'])
      partition_ipv6 = resource['address_list'][1]['addr']
      self.assertTrue(netaddr.valid_ipv6(partition_ipv6))
      for j, other_resource in enumerate(resource_list[i + 1:]):
        other_partition_ipv6 = other_resource['address_list'][1]['addr']
        self.assertNotEqual(partition_ipv6, addr0)
        if j % 2 == 1:
          self.assertEqual(partition_ipv6, other_partition_ipv6)
        else:
          self.assertNotEqual(partition_ipv6, other_partition_ipv6)

  def test_format_ipv6_slash_128_range(self):
    standalone = self.setupSimpleStandalone()
    prefixlen = 128
    slapos_fake_ipv6_range = '%s/%d' % (SLAPOS_TEST_IPV6, prefixlen)
    standalone.format(3, SLAPOS_TEST_IPV4, slapos_fake_ipv6_range)
    resource_list = self.getJsonResourceList(standalone)
    for i, resource in enumerate(resource_list):
      self.assertFalse(resource['ipv6_range'])
      partition_ipv6 = resource['address_list'][1]['addr']
      self.assertTrue(netaddr.valid_ipv6(partition_ipv6))
      self.assertEqual(partition_ipv6, SLAPOS_TEST_IPV6)

  def test_format_ipv6_no_range(self):
    standalone = self.setupSimpleStandalone()
    standalone.format(3, SLAPOS_TEST_IPV4, SLAPOS_TEST_IPV6)
    resource_list = self.getJsonResourceList(standalone)
    for i, resource in enumerate(resource_list):
      self.assertFalse(resource['ipv6_range'])
      partition_ipv6 = resource['address_list'][1]['addr']
      self.assertTrue(netaddr.valid_ipv6(partition_ipv6))
      for other_resource in resource_list[i + 1:]:
        self.assertEqual(partition_ipv6, SLAPOS_TEST_IPV6)

  def test_reformat_less_partitions(self):
    standalone = self.setupSimpleStandalone()
    standalone.format(2, SLAPOS_TEST_IPV4, SLAPOS_TEST_IPV6)
    standalone.format(1, SLAPOS_TEST_IPV4, SLAPOS_TEST_IPV6)
    self.assertNotExists(
        os.path.join(standalone.instance_directory, 'slappart1'))
    self.assertEqual(
        ['slappart0'],
        [cp.getId() for cp in standalone.computer.getComputerPartitionList()])

  def test_reformat_less_chmod_files(self):
    standalone = self.setupSimpleStandalone()
    standalone.format(2, SLAPOS_TEST_IPV4, SLAPOS_TEST_IPV6)
    # removing this directory should not be a problem
    chmoded_dir_path = os.path.join(standalone.instance_directory, 'slappart1', 'directory')
    os.mkdir(chmoded_dir_path)
    os.chmod(chmoded_dir_path, 0o000)
    standalone.format(1, SLAPOS_TEST_IPV4, SLAPOS_TEST_IPV6)
    self.assertNotExists(chmoded_dir_path)
    self.assertEqual(
        ['slappart0'],
        [cp.getId() for cp in standalone.computer.getComputerPartitionList()])

  def test_reformat_different_base_name(self):
    standalone = self.setupSimpleStandalone()
    standalone.format(
        1, SLAPOS_TEST_IPV4, SLAPOS_TEST_IPV6, partition_base_name="a")
    self.assertExists(os.path.join(standalone.instance_directory, 'a0'))
    standalone.format(
        1, SLAPOS_TEST_IPV4, SLAPOS_TEST_IPV6, partition_base_name="b")
    self.assertNotExists(os.path.join(standalone.instance_directory, 'a0'))
    self.assertExists(os.path.join(standalone.instance_directory, 'b0'))
    self.assertEqual(
        ['b0'],
        [cp.getId() for cp in standalone.computer.getComputerPartitionList()])

  def test_reformat_refuse_deleting_running_partition(self):
    standalone = self.setupSimpleStandalone()
    standalone.format(1, SLAPOS_TEST_IPV4, SLAPOS_TEST_IPV6)
    with mock.patch("slapos.slap.ComputerPartition.getState", return_value="busy"),\
       self.assertRaisesRegex(ValueError, "Cannot reformat to remove busy partition at .*slappart0"):
      standalone.format(0, SLAPOS_TEST_IPV4, SLAPOS_TEST_IPV6)

  def test_slapos_node_format(self):
    standalone = self.setupSimpleStandalone()
    self.assertExists(standalone.instance_directory)
    format_command = (standalone._slapos_wrapper, 'node', 'format', '--now')
    glob_pattern = os.path.join(standalone.instance_directory, 'slappart*')
    self.assertFalse(glob.glob(glob_pattern))
    self.assertTrue(subprocess.call(format_command)) # non-zero exitcode
    self.assertFalse(glob.glob(glob_pattern))
    for partition_count in (3, 2):
      standalone.format(partition_count, SLAPOS_TEST_IPV4, SLAPOS_TEST_IPV6)
      self.assertEqual(partition_count, len(glob.glob(glob_pattern)))
      subprocess.check_call(format_command)
      partitions = glob.glob(glob_pattern)
      self.assertEqual(partition_count, len(partitions))
      for path in partitions:
        slapos.util.rmtree(path)
      subprocess.check_call(format_command)
      self.assertEqual(partition_count, len(glob.glob(glob_pattern)))

  def test_two_instance_from_same_directory(self):
    working_dir = tempfile.mkdtemp(prefix=__name__)
    self.addCleanup(slapos.util.rmtree, working_dir)
    standalone1 = StandaloneSlapOS(
        working_dir, SLAPOS_TEST_IPV4, SLAPOS_TEST_PORT)

    def maybestop():
      # try to stop anyway, not to leak processes if test fail
      try:
        standalone1.stop()
      except:
        pass

    self.addCleanup(maybestop)

    # create another class instance, will control the same standanlone slapos.
    standalone2 = StandaloneSlapOS(
        working_dir, SLAPOS_TEST_IPV4, SLAPOS_TEST_PORT)
    standalone2.stop()

    # stopping standalone2 stops everything
    with self.assertRaises(BaseException):
      standalone1.supply("https://example.com/software.cfg")
    with self.assertRaises(BaseException):
      standalone1.stop()

  def test_partition_forward(self):
    # type: () -> None
    working_dir = tempfile.mkdtemp(prefix=__name__)
    self.addCleanup(slapos.util.rmtree, working_dir)
    partition_forward_config = [
        PartitionForwardConfiguration(
            'https://slapos1.example.com',
            'path/to/cert',
            'path/to/key',
            software_release_list=('https://example.com/software-1.cfg', ),
        ),
        PartitionForwardConfiguration(
            'https://slapos2.example.com',
            software_release_list=('https://example.com/software-2.cfg', ),
        ),
        PartitionForwardAsPartitionConfiguration(
            'https://slapos3.example.com',
            'computer',
            'partition',
            'path/to/cert',
            'path/to/key',
            software_release_list=('https://example.com/software-3.cfg', ),
        ),
        PartitionForwardAsPartitionConfiguration(
            'https://slapos4.example.com',
            'computer',
            'partition',
            software_release_list=('https://example.com/software-4.cfg', ),
        ),
    ]
    standalone = StandaloneSlapOS(
        working_dir,
        SLAPOS_TEST_IPV4,
        SLAPOS_TEST_PORT,
        partition_forward_configuration=partition_forward_config,
    )
    self.addCleanup(standalone.stop)

    config_parser = ConfigParser()
    config_parser.read([os.path.join(working_dir, 'etc', 'slapos.cfg')])
    self.assertTrue(
        config_parser.has_section('multimaster/https://slapos1.example.com'))
    self.assertEqual(
        'path/to/cert',
        config_parser.get('multimaster/https://slapos1.example.com', 'cert'))
    self.assertEqual(
        'path/to/key',
        config_parser.get('multimaster/https://slapos1.example.com', 'key'))
    self.assertEqual(
        'https://example.com/software-1.cfg',
        config_parser.get(
            'multimaster/https://slapos1.example.com',
            'software_release_list').strip())
    self.assertFalse(
        config_parser.has_option(
            'multimaster/https://slapos2.example.com', 'computer'))
    self.assertFalse(
        config_parser.has_option(
            'multimaster/https://slapos2.example.com', 'partition'))

    self.assertTrue(
        config_parser.has_section('multimaster/https://slapos2.example.com'))
    self.assertFalse(
        config_parser.has_option(
            'multimaster/https://slapos2.example.com', 'cert'))
    self.assertFalse(
        config_parser.has_option(
            'multimaster/https://slapos2.example.com', 'key'))

    self.assertTrue(
        config_parser.has_section('multimaster/https://slapos3.example.com'))
    self.assertEqual(
        'computer',
        config_parser.get(
            'multimaster/https://slapos3.example.com', 'computer'))
    self.assertEqual(
        'partition',
        config_parser.get(
            'multimaster/https://slapos3.example.com', 'partition'))

    self.assertTrue(
        config_parser.has_section('multimaster/https://slapos4.example.com'))


class SlapOSStandaloneTestCase(unittest.TestCase):
  # This test case takes care of stopping the standalone instance
  # in a cleanup, but subclasses who needs to control shutdown
  # can set this class attribute to False to prevent this behavior.
  _auto_stop_standalone = True

  def setUp(self):
    checkPortIsFree()
    working_dir = tempfile.mkdtemp(prefix=__name__)
    self.addCleanup(slapos.util.rmtree, working_dir)
    self.standalone = StandaloneSlapOS(
        working_dir,
        SLAPOS_TEST_IPV4,
        SLAPOS_TEST_PORT,
        shared_part_list=[
            os.path.expanduser(p) for p in os.environ.get(
                'SLAPOS_TEST_SHARED_PART_LIST', '').split(os.pathsep) if p
        ],
    )
    self.addCleanup(self.stopStandalone)
    self.standalone.format(1, SLAPOS_TEST_IPV4, SLAPOS_TEST_IPV6)

  def stopStandalone(self):
    if self._auto_stop_standalone:
      self.standalone.stop()
      self._auto_stop_standalone = False


class TestSlapOSStandaloneLogFile(SlapOSStandaloneTestCase):
  def test_log_files(self):
    log_directory = self.standalone._log_directory
    with open(os.path.join(log_directory, 'slapos-proxy.log')) as f:
      self.assertIn("Running on http", f.read())

    self.standalone.waitForSoftware()
    with open(os.path.join(log_directory, 'slapos-node-software.log')) as f:
      self.assertIn("Processing software releases", f.read())

    self.standalone.waitForInstance()
    with open(os.path.join(log_directory, 'slapos-node-instance.log')) as f:
      self.assertIn("Processing computer partitions", f.read())

    self.standalone.waitForReport()
    with open(os.path.join(log_directory, 'slapos-node-report.log')) as f:
      self.assertIn("Aggregating and sending usage reports", f.read())

    self.assertTrue(
        os.path.exists(
            os.path.join(log_directory, 'slapos-instance-supervisord.log')))


class TestSlapOSStandaloneSoftware(SlapOSStandaloneTestCase):
  def test_install_software(self):
    with tempfile.NamedTemporaryFile(suffix="-%s.cfg" % self.id()) as f:
      f.write(
          textwrap.dedent(
              '''
              [buildout]
              parts = instance
              newest = false
              [instance]
              recipe = plone.recipe.command==1.1
              command = touch ${buildout:directory}/instance.cfg
              update-command = touch ${buildout:directory}/updated
      ''').encode())
      f.flush()
      self.standalone.supply(f.name)
      self.standalone.waitForSoftware()

      software_hash = hashlib.md5(f.name.encode()).hexdigest()
      software_installation_path = os.path.join(
          self.standalone.software_directory, software_hash)
      self.assertTrue(os.path.exists(software_installation_path))
      self.assertTrue(
          os.path.exists(os.path.join(software_installation_path, 'bin')))
      self.assertTrue(
          os.path.exists(os.path.join(software_installation_path, 'parts')))
      self.assertTrue(
          os.path.exists(
              os.path.join(software_installation_path, '.completed')))
      self.assertTrue(
          os.path.exists(
              os.path.join(software_installation_path, 'instance.cfg')))

      # install respect the .completed file, once software is installed,
      # waitForSoftware will not process again software
      self.standalone.waitForSoftware()
      self.assertFalse(
          os.path.exists(
              os.path.join(software_installation_path, 'updated')))

      # waitForSoftware has a way to "force" reinstalling all software
      self.standalone.waitForSoftware(install_all=True)
      self.assertTrue(
          os.path.exists(
              os.path.join(software_installation_path, 'updated')))

      # destroy
      self.standalone.supply(f.name, state='destroyed')
      self.standalone.waitForSoftware()
      self.assertFalse(os.path.exists(software_installation_path))

  def test_install_software_failure(self):
    with tempfile.NamedTemporaryFile(suffix="-%s.cfg" % self.id()) as f:
      f.write(
          textwrap.dedent(
              '''
              [buildout]
              parts = error
              newest = false
              [error]
              recipe = plone.recipe.command==1.1
              command = bash -c "exit 123"
              stop-on-error = true
      ''').encode())
      f.flush()
      self.standalone.supply(f.name)

      with self.assertRaises(SlapOSNodeSoftwareError) as e:
        self.standalone.waitForSoftware()

      self.assertEqual(1, e.exception.args[0]['exitstatus'])
      self.assertIn(
          "Error: Non zero exit code (123) while running command.",
          e.exception.args[0]['output'])
      # SlapOSNodeSoftwareError.__str__ also displays output nicely
      self.assertIn(
          "SlapOSNodeSoftwareError exitstatus: 1 output:", str(e.exception))
      self.assertIn(
          "Error: Non zero exit code (123) while running command.",
          str(e.exception))
      self.assertNotIn(r"\n", str(e.exception))

  def test_install_software_failure_log_with_ansi_codes(self):
    with tempfile.NamedTemporaryFile(suffix="-%s.cfg" % self.id()) as f:
      f.write(
          textwrap.dedent(
              r'''
              [buildout]
              parts = error
              newest = false
              [error]
              recipe = plone.recipe.command==1.1
              command = bash -c "echo -e '\033[0;31mRed\033[0m \033[0;32mGreen\033[0m \033[0;34mBlue\033[0m' ; exit 123"
              stop-on-error = true
      ''').encode())
      f.flush()
      self.standalone.supply(f.name)

      with self.assertRaises(SlapOSNodeSoftwareError) as e:
        self.standalone.waitForSoftware()

      self.assertEqual(1, e.exception.args[0]['exitstatus'])
      self.assertIn("Red Green Blue", e.exception.args[0]['output'])


class TestSlapOSStandaloneInstance(SlapOSStandaloneTestCase):
  def test_request_instance(self):
    with tempfile.NamedTemporaryFile(suffix="-%s.cfg" % self.id()) as f:
      # This is a minimal / super fast buildout that's compatible with slapos.
      # We don't want to install slapos.cookbook because installation takes too
      # much time, so we use simple plone.recipe.command and shell.
      # This buildout create an instance with two parts:
      #  check_parameter: that checks that the requested parameter is set
      #  publish: that publish some parameters so that we can assert it's published.
      software_url = f.name
      f.write(
          textwrap.dedent(
              '''
              [buildout]
              parts = instance
              newest = false

              [instance]
              recipe = plone.recipe.command==1.1
              stop-on-error = true
              # we use @@DOLLAR@@{section:option} for what will become instance substitutions
              command = sed -e s/@@DOLLAR@@/$/g <<EOF > ${buildout:directory}/instance.cfg
                [buildout]
                parts = check_parameter publish
                eggs-directory = ${buildout:eggs-directory}

                [check_parameter]
                # check we were requested with request=parameter ( as a way to test
                # request parameters are sent )
                recipe = plone.recipe.command==1.1
                stop-on-error = true
                command = \\
                  curl '@@DOLLAR@@{slap-connection:server-url}/registerComputerPartition?computer_reference=@@DOLLAR@@{slap-connection:computer-id}&computer_partition_reference=@@DOLLAR@@{slap-connection:partition-id}' \\
                  | grep '<string>request</string><string>parameter</string>'

                [publish]
                # touch a file to check instance exists and publish a hardcoded parameter
                recipe = plone.recipe.command==1.1
                stop-on-error = true
                command = \\
                  touch instance.check \\
                    && curl -X POST @@DOLLAR@@{slap-connection:server-url}/setComputerPartitionConnectionXml \\
                    -d computer_id=@@DOLLAR@@{slap-connection:computer-id} \\
                    -d computer_partition_id=@@DOLLAR@@{slap-connection:partition-id} \\
                    -d connection_xml='<dictionary><string>published</string><string>parameter</string></dictionary>'
                EOF
      ''').encode())
      f.flush()

      self.standalone.supply(software_url)
      self.standalone.waitForSoftware()

    self.standalone.request(
        software_url,
        'default',
        'instance',
        partition_parameter_kw={'request': 'parameter'})
    self.standalone.waitForInstance()

    # check published parameters
    partition = self.standalone.request(
        software_url,
        'default',
        'instance',
        partition_parameter_kw={'request': 'parameter'})
    self.assertEqual({'published': 'parameter'},
                     partition.getConnectionParameterDict())

    # check instance files
    parition_directory = os.path.join(
        self.standalone.instance_directory, 'slappart0')
    self.assertTrue(
        os.path.exists(os.path.join(parition_directory, '.installed.cfg')))
    self.assertTrue(
        os.path.exists(os.path.join(parition_directory, 'instance.check')))

    # check instance supervisor, there should be only watchdog now
    with self.standalone.instance_supervisor_rpc as instance_supervisor_rpc:
      self.assertEqual(
          ['watchdog'],
          [p['name'] for p in instance_supervisor_rpc.getAllProcessInfo()])

    # delete instance
    self.standalone.request(
        software_url,
        'default',
        'instance',
        partition_parameter_kw={'partition': 'parameter'},
        state='destroyed',
    )
    self.standalone.waitForInstance()
    # instantiate does nothing, it will be deleted with `report`
    self.assertTrue(
        os.path.exists(os.path.join(parition_directory, 'instance.check')))
    self.standalone.waitForReport()
    self.assertFalse(
        os.path.exists(os.path.join(parition_directory, 'instance.check')))

    # check that stopping leaves no process
    process_list = []
    with self.standalone.instance_supervisor_rpc as instance_supervisor_rpc:
      process_list.append(psutil.Process(instance_supervisor_rpc.getPID()))
      process_list.extend([
          psutil.Process(p['pid'])
          for p in instance_supervisor_rpc.getAllProcessInfo()
          if p['statename'] == 'RUNNING'
      ])
    with self.standalone.system_supervisor_rpc as system_supervisor_rpc:
      process_list.append(psutil.Process(system_supervisor_rpc.getPID()))
      process_list.extend([
          psutil.Process(p['pid'])
          for p in system_supervisor_rpc.getAllProcessInfo()
          if p['statename'] == 'RUNNING'
      ])
    self.assertEqual(set([True]), set([p.is_running() for p in process_list]))
    self.stopStandalone()
    self.assertEqual(set([False]), set([p.is_running() for p in process_list]))


class TestSlapOSStandaloneLongSoftwareError(SlapOSStandaloneTestCase):
  def test(self):
    with tempfile.NamedTemporaryFile(suffix="-%s.cfg" % self.id()) as f:
      very_long_string = "a" * (2 << 15)
      software_url = f.name
      f.write(
          textwrap.dedent(
              '''
              [buildout]
              parts = fail
              newest = false

              [fail]
              recipe = plone.recipe.command==1.1
              stop-on-error = true
              command = echo %r && false
      ''' % very_long_string).encode())
      f.flush()

      self.standalone.supply(software_url)

      # Not using assertRaises context manager for python2 compatibility
      try:
        self.standalone.waitForSoftware()
      except SlapOSNodeSoftwareError as e:
        self.assertIn(very_long_string, str(e))
      except Exception as e:
        self.fail("waitForSoftware raised unexpected exception %r" % e)
      else:
        self.fail("waitForSoftware should have raised")